# Utility functions
#
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the Util module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'aymgal'


import copy
import numpy as np
# import scipy.ndimage.interpolation as interp
import herculens.Util.util as util
from herculens.Util import image_util
from herculens.LightModel.Profiles.gaussian import Gaussian


def kernel_norm(kernel):
    """

    :param kernel:
    :return: normalisation of the psf kernel
    """
    return kernel / np.sum(kernel)

def subgrid_kernel(kernel, subgrid_res, odd=False, num_iter=100):
    """
    creates a higher resolution kernel with subgrid resolution as an interpolation of the original kernel in an
    iterative approach

    :param kernel: initial kernel
    :param subgrid_res: subgrid resolution required
    :return: kernel with higher resolution (larger)
    """
    subgrid_res = int(subgrid_res)
    if subgrid_res == 1:
        return kernel
    nx, ny = np.shape(kernel)
    d_x = 1. / nx
    x_in = np.linspace(d_x/2, 1-d_x/2, nx)
    d_y = 1. / nx
    y_in = np.linspace(d_y/2, 1-d_y/2, ny)
    nx_new = nx * subgrid_res
    ny_new = ny * subgrid_res
    if odd is True:
        if nx_new % 2 == 0:
            nx_new -= 1
        if ny_new % 2 == 0:
            ny_new -= 1

    d_x_new = 1. / nx_new
    d_y_new = 1. / ny_new
    x_out = np.linspace(d_x_new/2., 1-d_x_new/2., nx_new)
    y_out = np.linspace(d_y_new/2., 1-d_y_new/2., ny_new)
    kernel_input = copy.deepcopy(kernel)
    kernel_subgrid = image_util.re_size_array(x_in, y_in, kernel_input, x_out, y_out)
    kernel_subgrid = kernel_norm(kernel_subgrid)
    for i in range(num_iter):
        # given a proposition, re-size it to original pixel size
        if subgrid_res % 2 == 0:
            kernel_pixel = averaging_even_kernel(kernel_subgrid, subgrid_res)
        else:
            kernel_pixel = util.averaging(kernel_subgrid, numGrid=nx_new, numPix=nx)
        delta = kernel - kernel_pixel
        temp_kernel = kernel_input + delta
        kernel_subgrid = image_util.re_size_array(x_in, y_in, temp_kernel, x_out, y_out)#/norm_subgrid
        kernel_subgrid = kernel_norm(kernel_subgrid)
        kernel_input = temp_kernel

    #from scipy.ndimage import zoom

    #ratio = subgrid_res
    #kernel_subgrid = zoom(kernel, ratio, order=4) / ratio ** 2
    #print(np.shape(kernel_subgrid))
    # whatever has not been matched is added to zeroth order (in squares of the undersampled PSF)
    if subgrid_res % 2 == 0:
        return kernel_subgrid
    kernel_pixel = util.averaging(kernel_subgrid, numGrid=nx_new, numPix=nx)
    kernel_pixel = kernel_norm(kernel_pixel)
    delta_kernel = kernel_pixel - kernel_norm(kernel)
    id = np.ones((subgrid_res, subgrid_res))
    delta_kernel_sub = np.kron(delta_kernel, id)/subgrid_res**2
    return kernel_norm(kernel_subgrid - delta_kernel_sub)

def averaging_even_kernel(kernel_high_res, subgrid_res):
    """
    makes a lower resolution kernel based on the kernel_high_res (odd numbers) and the subgrid_res (even number), both
    meant to be centered.

    :param kernel_high_res: high resolution kernel with even subsampling resolution, centered
    :param subgrid_res: subsampling resolution (even number)
    :return: averaged undersampling kernel
    """
    n_kernel_high_res = len(kernel_high_res)
    n_low = int(round(n_kernel_high_res / subgrid_res + 0.5))
    if n_low % 2 == 0:
        n_low += 1
    n_high = int(n_low * subgrid_res - 1)
    assert n_high % 2 == 1
    if n_high == n_kernel_high_res:
        kernel_high_res_edges = kernel_high_res
    else:
        i_start = int((n_high - n_kernel_high_res) / 2)
        kernel_high_res_edges = np.zeros((n_high, n_high))
        kernel_high_res_edges[i_start:-i_start, i_start:-i_start] = kernel_high_res
    kernel_low_res = np.zeros((n_low, n_low))
    # adding pixels that are fully within a single re-binned pixel
    for i in range(subgrid_res-1):
        for j in range(subgrid_res-1):
            kernel_low_res += kernel_high_res_edges[i::subgrid_res, j::subgrid_res]
    # adding half of a pixel that has over-lap with two pixels
    i = subgrid_res - 1
    for j in range(subgrid_res - 1):
        kernel_low_res[1:, :] += kernel_high_res_edges[i::subgrid_res, j::subgrid_res] / 2
        kernel_low_res[:-1, :] += kernel_high_res_edges[i::subgrid_res, j::subgrid_res] / 2
    j = subgrid_res - 1
    for i in range(subgrid_res - 1):
        kernel_low_res[:, 1:] += kernel_high_res_edges[i::subgrid_res, j::subgrid_res] / 2
        kernel_low_res[:, :-1] += kernel_high_res_edges[i::subgrid_res, j::subgrid_res] / 2
    # adding a quarter of a pixel value that is at the boarder of four pixels
    i = subgrid_res - 1
    j = subgrid_res - 1
    kernel_edge = kernel_high_res_edges[i::subgrid_res, j::subgrid_res]
    kernel_low_res[1:, 1:] += kernel_edge / 4
    kernel_low_res[:-1, 1:] += kernel_edge / 4
    kernel_low_res[1:, :-1] += kernel_edge / 4
    kernel_low_res[:-1, :-1] += kernel_edge / 4
    return kernel_low_res

def cut_psf(psf_data, psf_size):
    """
    cut the psf properly
    :param psf_data: image of PSF
    :param psf_size: size of psf
    :return: re-sized and re-normalized PSF
    """
    kernel = image_util.cut_edges(psf_data, psf_size)
    kernel = kernel_norm(kernel)
    return kernel

def kernel_gaussian(kernel_numPix, deltaPix, fwhm):
    sigma = util.fwhm2sigma(fwhm)
    #if kernel_numPix % 2 == 0:
    #    kernel_numPix += 1
    x_grid, y_grid = util.make_grid(kernel_numPix, deltaPix)
    gaussian = Gaussian()
    kernel = gaussian.function(x_grid, y_grid, amp=1., sigma=sigma, center_x=0, center_y=0)
    kernel /= np.sum(kernel)
    return kernel.reshape(kernel_numPix, kernel_numPix)

def split_kernel(kernel_super, supersampling_kernel_size, supersampling_factor, normalized=True):
    """
    pixel kernel and subsampling kernel such that the convolution of both applied on an image can be
    performed, i.e. smaller subsampling PSF and hole in larger PSF

    :param kernel: PSF kernel of the size of the pixel
    :param kernel_super: super-sampled kernel
    :param supersampling_kernel_size: size of super-sampled PSF in units of degraded pixels
    :param normalized: boolean, if True returns a split kernel that is area normalized=1 representing a convolution kernel
    :return: degraded kernel with hole and super-sampled kernel
    """
    if supersampling_factor <= 1:
        raise ValueError('To split a kernel, the supersampling_factor needs to be > 1, givn %s' %supersampling_factor)
    if supersampling_kernel_size % 2 == 0:
        raise ValueError('supersampling_kernel_size needs to be an odd number!')
    n_super = len(kernel_super)
    n_sub = supersampling_kernel_size * supersampling_factor
    if n_sub % 2 == 0:
        n_sub += 1
    if n_sub > n_super:
        n_sub = n_super

    kernel_hole = copy.deepcopy(kernel_super)
    n_min = int((n_super-1) / 2 - (n_sub - 1) / 2)
    n_max = int((n_super-1) / 2 + (n_sub - 1) / 2 + 1)
    kernel_hole[n_min:n_max, n_min:n_max] = 0
    kernel_hole_resized = degrade_kernel(kernel_hole, degrading_factor=supersampling_factor)
    kernel_subgrid_cut = kernel_super[n_min:n_max, n_min:n_max]
    if normalized is True:
        flux_subsampled = np.sum(kernel_subgrid_cut)
        flux_hole = np.sum(kernel_hole_resized)
        if flux_hole > 0:
            kernel_hole_resized *= (1. - flux_subsampled) / np.sum(kernel_hole_resized)
        else:
            kernel_subgrid_cut /= np.sum(kernel_subgrid_cut)
    else:
        kernel_hole_resized /= supersampling_factor ** 2
    return kernel_hole_resized, kernel_subgrid_cut

def degrade_kernel(kernel_super, degrading_factor):
    """

    :param kernel_super: higher resolution kernel (odd number per axis)
    :param degrading_factor: degrading factor (effectively the super-sampling resolution of the kernel given
    :return: degraded kernel with odd axis number with the sum of the flux/values in the kernel being preserved
    """
    if degrading_factor == 1:
        return kernel_super
    if degrading_factor % 2 == 0:
        kernel_low_res = averaging_even_kernel(kernel_super, degrading_factor)
    else:
        n_kernel = len(kernel_super)
        numPix = int(round(n_kernel / degrading_factor + 0.5))
        if numPix % 2 == 0:
            numPix += 1
        n_high = numPix * degrading_factor

        kernel_super_ = np.zeros((n_high, n_high))
        i_start = int((n_high-n_kernel)/2)
        kernel_super_[i_start:i_start+n_kernel, i_start:i_start+n_kernel] = kernel_super
        kernel_low_res = util.averaging(kernel_super_, numGrid=n_high, numPix=numPix) * degrading_factor**2  # multiplicative factor added when providing flux conservation
    return kernel_low_res

def fwhm_kernel(kernel):
    """

    :param kernel:
    :return:
    """
    n = len(kernel)
    center = (n - 1) / 2.
    I_r = image_util.radial_profile(kernel, center=[center, center])
    if n % 2 == 0:
        raise ValueError('only works with odd number of pixels in kernel!')
    max_flux = kernel[int((n-1)/2), int((n-1)/2)]
    I_2 = max_flux / 2.
    r = np.linspace(0, (n - 1) / 2, int((n + 1) / 2)) + 0.33
    for i in range(1, len(I_r)):
        if I_r[i] < I_2:
            fwhm_2 = (I_2 - I_r[i - 1]) / (I_r[i] - I_r[i - 1]) + r[i - 1]
            return fwhm_2 * 2
    raise ValueError('The kernel did not drop to half the max value - fwhm not determined!')



# Classes and functions to use with JAX
#
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'austinpeel', 'aymgal', 'duxfrederic'


from copy import deepcopy
import numpy as np
import jax.numpy as jnp
from jax import jit, lax
from jax.scipy.special import gammaln


def unjaxify_kwargs(kwargs_params):
    """
    Utility to convert all JAX's device arrays contained in a model kwargs
    to standard floating point or numpy arrays.
    """
    def unjaxify_array(p):
        if p is None or isinstance(p, (float, int)):
            return p  # don't do anything
        if p.size == 1:
            return float(p)
        return np.array(p)
    kwargs_params_new = deepcopy(kwargs_params)
    for model_key, model_kwargs in kwargs_params.items():
        for profile_idx, profile_kwargs in enumerate(model_kwargs):
            for k, p in profile_kwargs.items():
                if isinstance(p, (tuple, list)):
                    # iterate over the list/tuple and unjaxify the items
                    kwargs_params_new[model_key][profile_idx][k] = []  # NOTE: if it was a tuple before, it's not anymore
                    for p_sub in p:
                        kwargs_params_new[model_key][profile_idx][k].append(unjaxify_array(p_sub))
                else:
                    # unjaxify the array
                    kwargs_params_new[model_key][profile_idx][k] = unjaxify_array(p)
    return kwargs_params_new


def R_omega(z, t, q, nmax):
    """Angular dependency of the deflection angle in the EPL lens profile.

    The computation follows Eqs. (22)-(29) in Tessore & Metcalf (2015), where
    z = R * e^(i * phi) is a position vector in the lens plane,
    t = gamma - 1 is the logarithmic slope of the profile, and
    q is the axis ratio.

    This iterative implementation is necessary, since the usual hypergeometric
    function `hyp2f1` provided in `scipy.special` has not yet been implemented
    in an autodiff way in JAX.

    Note that the value returned includes an extra factor R multiplying Eq. (23)
    for omega(phi).

    """
    # Compute constant factors
    f = (1. - q) / (1. + q)
    ei2phi = z / z.conjugate()
    # Set the first term of the series
    omega_i = z  # jnp.array(np.copy(z))  # Avoid overwriting z ?
    partial_sum = omega_i

    @jit
    def body_fun(i, val):
        # Current term in the series is proportional to the previous
        ratio = (2. * i + t - 2.) / (2. * i - t + 2.)
        val[1] = -f * ei2phi * ratio * val[1]
        # Adds the current term to the partial sum
        val[0] += val[1]
        return val

    return lax.fori_loop(1, nmax, body_fun, [partial_sum, omega_i])[0]


def omega_real(x, y, t, q, nmax):
    """Angular dependency of the deflection angle in the EPL lens profile.

    The computation follows Eqs. (31)-(32) in Tessore & Metcalf (2015), where
    x, y are coordinates in the lens plane,
    t = gamma - 1 is the logarithmic slope of the profile, and
    q is the axis ratio.

    This iterative implementation is necessary, since the usual hypergeometric
    function `hyp2f1` provided in `scipy.special` has not yet been implemented
    in an autodiff way in JAX.

    This function is based on the Giga-lens implementation (gigalens.jax.profiles.mass.epl).
    """
    # Compute constant factors
    phi = jnp.arctan2(y, q * x)
    f = (1. - q) / (1. + q)
    Cs, Ss = jnp.cos(phi), jnp.sin(phi)
    Cs2, Ss2 = jnp.cos(2 * phi), jnp.sin(2 * phi)
    def update(n, val):
        prefac = -f * (2. * n - (2. - t)) / (2. * n + (2. - t))
        last_x, last_y, fx, fy = val
        last_x, last_y = prefac * (Cs2 * last_x - Ss2 * last_y), prefac * (
                Ss2 * last_x + Cs2 * last_y
        )
        fx += last_x
        fy += last_y
        return last_x, last_y, fx, fy
    _, _, fx, fy = lax.fori_loop(1, nmax, update, (Cs, Ss, Cs, Ss))
    return fx, fy


class special(object):
    @staticmethod
    @jit
    def gamma(x):
        """Gamma function.

        This function is necessary in many lens mass models, but JAX does not
        currently have an implementation in jax.scipy.special. Values of the
        function are computed simply as the exponential of the logarithm of the
        gamma function (which has been implemented in jax), taking the sign
        of negative inputs properly into account.

        Note that even when just-in-time compiled, this function is much
        slower than its original scipy counterpart.

        Parameters
        ----------
        x : array_like
            Real-valued argument.

        Returns
        -------
        scalar or ndarray
            Values of the Gamma function.
        """
        # Properly account for the sign of negative x values
        sign_condition = (x > 0) | (jnp.ceil(x) % 2 != 0) | (x % 2 == 0)
        sign = 2 * jnp.asarray(sign_condition, dtype=float) - 1
        return sign * jnp.exp(gammaln(x))



# Defines the model of a strong lens
# 
# Copyright (c) 2023, herculens developers and contributors
# Copyright (c) 2020, Austin Peel and Aymeric Galan
# based on the LensingOperator class from slitronomy (https://github.com/aymgal/SLITronomy)

__author__ = 'austinpeel', 'aymgal'


import copy
import numpy as np
import jax.numpy as jnp
from functools import partial
from jax import jit
from jax.experimental import sparse as jsparse

from herculens.Util import util


__all__ = ['LensingOperator', 'PlaneGrid', 'MaskedPlaneGrid']


class LensingOperator(object):

    """Defines the mapping of pixelated light profiles between image and source planes"""

    def __init__(self, mass_model_class, image_grid_class, source_grid_class, arc_mask=None):
        """Summary
        
        Parameters
        ----------
        mass_model_class : TYPE
            Description
        image_grid_class : TYPE
            Description
        source_grid_class : TYPE
            Description
        
        Raises
        ------
        ValueError
            Description
        """
        self.MassModel = mass_model_class
        self.ImagePlane = MaskedPlaneGrid(image_grid_class, mask=arc_mask)
        self.SourcePlane = PlaneGrid(source_grid_class)

    @partial(jit, static_argnums=(0,))
    def source2image(self, source_1d):
        image = self._mapping @ source_1d
        return self.ImagePlane.place(image)

    @partial(jit, static_argnums=(0,))
    def source2image_2d(self, source):
        source_1d = util.image2array(source)
        return util.array2image(self.source2image(source_1d))

    @partial(jit, static_argnums=(0, 2))
    def image2source(self, image_1d, no_flux_norm=False):
        """if no_flux_norm is True, do not normalize light flux to better visualize the mapping"""
        source = self._mapping.T @ self.ImagePlane.extract(image_1d)
        if not no_flux_norm:
            source /= self._norm_image2source
        return source

    @partial(jit, static_argnums=(0, 2))
    def image2source_2d(self, image, no_flux_norm=False):
        image_1d = util.image2array(image)
        return util.array2image(self.image2source(image_1d, no_flux_norm=no_flux_norm))

    @partial(jit, static_argnums=(0,))
    def lensing(self, source):
        return self.source2image_2d(source)

    @partial(jit, static_argnums=(0,))
    def lensing_transpose(self, image):
        return self.image2source_2d(image)

    @property
    def source_plane_coordinates(self):
        return self.SourcePlane.theta_x, self.SourcePlane.theta_y

    @property
    def image_plane_coordinates(self):
        return self.ImagePlane.theta_x, self.ImagePlane.theta_y

    @property
    def pixel_area_ratio(self):
        """source pixel area divide by image pixel area"""
        return (self.SourcePlane.delta_pix / self.ImagePlane.delta_pix)**2

    def get_lens_mapping(self):
        if not hasattr(self, '_mapping'):
            return None
        return self._mapping, self._norm_image2source

    def delete_cache(self):
        if hasattr(self, '_mapping'):
            del self._mapping
        if hasattr(self, '_norm_image2source'):
            del self._norm_image2source

    def compute_mapping(self, kwargs_lens):
        # delete cached mapping matrices
        self.delete_cache()

        # compute mapping between image and source planes due to lensing, on original source plane grid
        self._mapping, self._norm_image2source = self._compute_mapping(kwargs_lens)

    def _compute_mapping(self, kwargs_lens):
        return self._compute_mapping_bilinear(kwargs_lens)

    def _compute_mapping_bilinear(self, kwargs_lens, resized_source_plane=True):
        """Compute the mapping between image and source plane pixels.

        This method ray-traces the image plane pixel
        coordinates back to the source plane and regularizes the resultig
        positions to a grid. In contrast to the 'nearest' interpolation scheme,
        this mapping incorporates a bilinear weighting to interpolate flux on
        the source plane following Treu & Koopmans (2004).

        """
        # Get image plane coordinates
        theta_x, theta_y = self.image_plane_coordinates

        # Compute lens mapping from image to source coordinates
        beta_x, beta_y = self.MassModel.ray_shooting(theta_x, theta_y, kwargs_lens)

        # Determine source pixels and their appropriate weights
        indices, weights = self._find_source_pixels_bilinear(beta_x, beta_y, 0., 0.)
    
        # Build lensing matrix as a sparse matrix for saving memory
        dense_shape = (self.ImagePlane.grid_size, self.SourcePlane.grid_size)
        lens_mapping = jsparse.BCOO((weights, np.stack(indices, axis=1)), 
                                    shape=dense_shape, indices_sorted=True)

        # Compue the flux normalization factors
        norm_image2source = np.maximum(1, lens_mapping.sum(axis=0).todense())
        return lens_mapping, norm_image2source

    def _find_source_pixels_bilinear(self, beta_x, beta_y, grid_offset_x, grid_offset_y, 
                                     warning=False):
        """Fast binning of ray-traced coordinates and weight calculation.

        Parameters
        ----------
        beta_x, beta_y : array-like
            Coordinates in the source plane of ray-traced points from the
            image plane.
        grid_offset_x, grid_offset_y : float
            Amount by which to shift the source plane grid in each direction.
        warning : bool
            Print a warning if any returned weights are negative.

        Returns
        -------
        (row, col), weight
            Weights are the source grid interpolation values, which belong
            at position (row, col) in the sparse lensing matrix. There are at
            most 4 weights corresponding to each row.

        Notes
        -----
        Ray-traced coordinates from the image plane are simply removed if they
        fall outside of the source plane grid, as is done in Treu & Koopmans
        (2004). Although this should only rarely occur in practice, e.g. for
        extreme parameters of the lens model, a better approach might still be
        to expand the source plane instead.

        """
        # Standardize inputs for vectorization
        beta_x = np.atleast_1d(beta_x)
        beta_y = np.atleast_1d(beta_y)
        assert len(beta_x) == len(beta_y), "Input arrays must be the same size."
        num_beta = len(beta_x)

        # Get source plane coordinates, shift them if necessary
        source_theta_x, source_theta_y = self.source_plane_coordinates
        source_theta_x += grid_offset_x
        source_theta_y += grid_offset_y

        # Compute bin edges so that (theta_x, theta_y) lie at the grid centers
        num_pix = self.SourcePlane.num_pix
        delta_pix = self.SourcePlane.delta_pix
        half_pix = delta_pix / 2

        theta_x = source_theta_x[:num_pix]
        x_dir = -1 if theta_x[0] > theta_x[-1] else 1  # Handle x-axis inversion
        # x_dir = lax.cond(theta_x[0] > theta_x[-1], lambda _: -1, lambda _: +1, None)  # Handle x-axis inversion
        x_lower = theta_x[0] - x_dir * half_pix
        x_upper = theta_x[-1] + x_dir * half_pix
        xbins = np.linspace(x_lower, x_upper, num_pix + 1)

        theta_y = source_theta_y[::num_pix]
        y_dir = -1 if theta_y[0] > theta_y[-1] else 1  # Handle y-axis inversion
        # y_dir = lax.cond(theta_y[0] > theta_y[-1], lambda _: -1, lambda _: +1, None)  # Handle y-axis inversion
        y_lower = theta_y[0] - y_dir * half_pix
        y_upper = theta_y[-1] + y_dir * half_pix
        ybins = np.linspace(y_lower, y_upper, num_pix + 1)

        # Keep only betas that fall within the source plane grid
        x_min, x_max = [x_lower, x_upper][::x_dir]
        y_min, y_max = [y_lower, y_upper][::y_dir]
        # x_min, x_max = lax.cond(x_dir == 1, lambda _: (x_lower, x_upper), lambda _: (x_upper, x_lower), None)
        # y_min, y_max = lax.cond(y_dir == 1, lambda _: (y_lower, y_upper), lambda _: (y_upper, y_lower), None)
        selection = ((beta_x > x_min) & (beta_x < x_max) &
                     (beta_y > y_min) & (beta_y < y_max))
        if np.any(1 - selection.astype(int)):
           beta_x = beta_x[selection]
           beta_y = beta_y[selection]
           num_beta = len(beta_x)

        # Find the (1D) source plane pixel that (beta_x, beta_y) falls in
        index_x = np.digitize(beta_x, xbins) - 1
        index_y = np.digitize(beta_y, ybins) - 1
        index_1 = index_x + index_y * num_pix

        # Compute distances between ray-traced betas and source grid points
        dx = beta_x - source_theta_x[index_1]
        dy = beta_y - source_theta_y[index_1]

        # Find the three other nearest pixels (may end up out of bounds)
        index_2 = index_1 + x_dir * np.sign(dx).astype(int)
        index_3 = index_1 + y_dir * np.sign(dy).astype(int) * num_pix
        index_4 = index_2 + y_dir * np.sign(dy).astype(int) * num_pix

        # Treat these index arrays as four sets stacked vertically
        # Prepare to mask out out-of-bounds pixels as well as repeats
        # The former is important for the csr_matrix to be generated correctly
        max_index = self.SourcePlane.grid_size - 1  # Upper index bound
        mask = np.ones((4, num_beta), dtype=bool)  # Mask for the betas

        # Mask out any neighboring pixels that end up out of bounds
        mask[1, np.where((index_2 < 0) | (index_2 > max_index))[0]] = False
        mask[2, np.where((index_3 < 0) | (index_3 > max_index))[0]] = False
        mask[3, np.where((index_4 < 0) | (index_4 > max_index))[0]] = False
        
        # Mask any repeated pixels (2 or 3x) arising from unlucky grid alignment
        zero_dx = np.where(dx == 0)[0]
        zero_dy = np.where(dy == 0)[0]
        unique, counts = np.unique(zero_dx + zero_dy, return_counts=True)
        repeat_row = [ii + 1 for c in counts for ii in range(0, 3, 3 - c)]
        repeat_col = [u for (u, c) in zip(unique, counts) for _ in range(c + 1)]
        mask[(repeat_row, repeat_col)] = False
        # mask.at[(repeat_row, repeat_col)].set(False)

        # Generate 2D indices of non-zero elements for the sparse matrix
        row = np.tile(np.nonzero(selection)[0], (4, 1))
        col = np.array([index_1, index_2, index_3, index_4])

        # Compute bilinear weights like in Treu & Koopmans (2004)
        col[~mask] = 0
        # col.at[~mask].set(0)  # Avoid accessing source_thetas out of bounds
        dist_x = (np.tile(beta_x, (4, 1)) - source_theta_x[col]) / delta_pix
        dist_y = (np.tile(beta_y, (4, 1)) - source_theta_y[col]) / delta_pix
        weight = (1 - np.abs(dist_x)) * (1 - np.abs(dist_y))

        # Make sure the weights are properly normalized
        # This step is only necessary where the mask has excluded source pixels
        norm = np.expand_dims(np.sum(weight, axis=0, where=mask), 0)
        weight = weight / norm

        if warning:
            if np.any(weight[mask] < 0):
                num_neg = np.sum((weight[mask] < 0).astype(int))
                print("Warning : {} weights are negative.".format(num_neg))

        return (row[mask], col[mask]), weight[mask]


class PlaneGrid(object):

    """
    Base class for image and source plane grids, designed for pixelated lensing operator.
    """

    def __init__(self, grid_class):
        """Initialise the grid.
        
        Parameters
        ----------
        grid_class : herculens.Coordinates.pixel_grid.PixelGrid
            PixelGrid instance
        """
        x_grid_2d, y_grid_2d = grid_class.pixel_coordinates
        self._x_grid_1d, self._y_grid_1d = util.image2array(x_grid_2d), util.image2array(y_grid_2d)
        self._delta_pix = grid_class.pixel_width
        num_pix_x, num_pix_y = grid_class.num_pixel_axes
        if num_pix_x != num_pix_y:
            raise ValueError("Only square images are supported")
        self._num_pix = num_pix_x
        self._subgrid_res = 1. # grid_class.supersampling_factor

    @property
    def num_pix(self):
        return self._num_pix

    @property
    def grid_size(self):
        return self.num_pix**2

    @property
    def grid_shape(self):
        return (self._num_pix, self._num_pix)

    @property
    def delta_pix(self):
        return self._delta_pix

    @property
    def theta_x(self):
        return self._x_grid_1d

    @property
    def theta_y(self):
        return self._y_grid_1d

    @property
    def subgrid_resolution(self):
        return self._subgrid_res


class MaskedPlaneGrid(PlaneGrid):

    def __init__(self, grid_class, mask=None):
        super(MaskedPlaneGrid, self).__init__(grid_class)
        if mask is None:
            self._no_mask = True
            self._num_pix_eff = self._num_pix**2
            # self._mask_1d = np.ones(self._num_pix**2).astype(bool)
        else:
            self._no_mask = False
            self._mask_1d = mask.ravel().astype(bool)
            self._num_pix_eff = np.count_nonzero(self._mask_1d)

    @property
    def theta_x(self):
        return self._x_grid_1d if self._no_mask else self._x_grid_1d[self._mask_1d]

    @property
    def theta_y(self):
        return self._y_grid_1d if self._no_mask else self._y_grid_1d[self._mask_1d]

    @property
    def num_pix(self):
        raise self._num_pix if self._no_mask else None

    @property
    def grid_size(self):
        return self._num_pix_eff

    def place(self, values):
        if self._no_mask:
            return values
        image = jnp.zeros(self._num_pix**2)
        return image.at[self._mask_1d].set(values)

    def extract(self, values):
        if self._no_mask:
            return values
        return values[self._mask_1d]
        


# Defines the model of a strong lens
# 
# Copyright (c) 2022, herculens developers and contributors


__author__ = 'aymgal'


import copy
import jax
import jax.numpy as jnp
import numpyro
from numpyro import handlers
from numpyro.infer import util

from herculens.Inference.ProbModel.base_model import BaseProbModel
from herculens.Util import numpyro_util as my_util


__all__ = ['NumpyroModel']


class NumpyroModel(BaseProbModel):
    """Defines a numpyro model based on a LensImage instance"""

    @property
    def num_parameters(self):
        if not hasattr(self, '_num_param'):
            num_param = 0
            prng_key = jax.random.PRNGKey(0)  # seed not relevant
            for site in self.get_trace(prng_key).values():
                if (site['type'] == 'sample' and not site['is_observed']
                    or site['type'] == 'param'):
                    num_param += site['value'].size
            self._num_param = num_param
        return self._num_param

    def log_prob(self, params, constrained=False):
        """returns the logarithm of the data likelihood plus the logarithm of the prior"""
        if constrained is True:
            # do this for optimisation in constrained space
            log_prob, model_trace = util.log_density(self.model, (), {}, params)
        else:
            # do this for optimisation in unconstrained space
            # TODO: use the new numpyro function potential_fn instead
            log_prob = - my_util.potential_energy(self.model, (), {}, params)
        return log_prob
    
    def log_likelihood(self, params, obs_site_key='obs'):
        # returns the logarithm of the data likelihood
        return util.log_likelihood(self.model, params, batch_ndims=0)[obs_site_key]

    def seeded_model(self, prng_key):
        return handlers.seed(self.model, prng_key)
    
    def get_trace(self, prng_key):
        return handlers.trace(self.seeded_model(prng_key)).get_trace()

    def get_sample(self, prng_key=None):
        if prng_key is None:
            prng_key = jax.random.PRNGKey(0)
        trace = self.get_trace(prng_key)
        return {site['name']: site['value'] for site in trace.values() if not site.get('is_observed', False)}

    def sample_prior(self, num_samples, prng_key=None):
        if prng_key is None:
            prng_key = jax.random.PRNGKey(0)
        batch_ndims = 0 if num_samples else 1
        predictive = util.Predictive(self.model, 
                                     num_samples=num_samples, 
                                     batch_ndims=batch_ndims)
        samples = predictive(prng_key)
        # delete all sites whose key contain 'obs'
        sites_keys = copy.deepcopy(list(samples.keys()))
        for key in sites_keys:
            if 'obs' in key:
                del samples[key]
        return samples

    def render_model(self):
        return numpyro.render_model(self.model)

    def constrain(self, params):
        return util.constrain_fn(self.model, (), {}, params)

    def unconstrain(self, params):
        return util.unconstrain_fn(self.model, (), {}, params)



__all__ = ['BaseRegularization']


class BaseRegularization(object):
    """Base class for all regularization methods."""

    param_names = []
    lower_limit_default = {}
    upper_limit_default = {}
    fixed_default = {}

    def __init__(self, model_type, profile_index, mass_form=None):
        if model_type not in ['source', 'lens_light', 'mass_light']:
            raise ValueError("Unsupported model type for regularization")
        self.model_type = model_type
        self.profile_idx = profile_index
        if self.model_type == 'source':
            self._param_key = 'kwargs_source'
        elif self.model_type == 'lens_light':
            self._param_key = 'kwargs_lens_light'
        elif self.model_type == 'lens_mass':
            self._param_key = 'kwargs_lens'
        self._mass_form = mass_form

    def log_prob(self, kwargs_hyperparams, kwargs_params):
        raise NotImplementedError

    def initialize(self, *args, **kwargs):
        raise NotImplementedError

    def update(self, *args, **kwargs):
        raise NotImplementedError

    def get_pixel_params(self, kwargs_params):
        return kwargs_params[self._param_key][self.profile_idx]['pixels']

    @property
    def has_weights(self):
        # some regularization methods require weights
        return hasattr(self, 'weights')



# Copyright (c) 2023, herculens developers and contributors

__author__ = 'austinpeel'

import functools
import numpy as np
import jax.numpy as jnp

try:
    from helens import LensEquationSolver
except ImportError:
    _solver_installed = False
else:
    _solver_installed = True


__all__ = ['PointSource']


class PointSource(object):
    """A point source defined in the image or source plane.

    A point source is considered to be either
    (1) a single position and amplitude defined in the source plane, or else
    (2) multiple positions and amplitudes defined in the image plane which
        correspond to a single point in the source plane.

    """

    def __init__(self, point_source_type, mass_model, image_plane):
        """Instantiate a point source.

        Parameters
        ----------
        point_source_type : str
            Either 'LENSED_POSITIONS' or 'SOURCE_POSITION'.
        mass_model : instance of `herculens.MassModel.mass_model.MassModel`
            Model of the lensing mass used to map positions between the source
            and image planes. Default is None.
        image_plane : instance of `herculens.Coordinates.pixel_grid.PixelGrid`
            Pixel grid used for triangulation in solving the lens equation.

        """
        self.type = point_source_type
        self.mass_model = mass_model
        self.image_plane = image_plane
        if self.type == 'SOURCE_POSITION':
            self._check_solver_install(f"type = '{self.type}'")

    @property
    def solver(self):
        if not hasattr(self, '_solver'):
            # TODO: support the argument k != None
            ray_shooting_func = functools.partial(self.mass_model.ray_shooting, k=None)
            x_grid, y_grid = self.image_plane.pixel_coordinates
            self._solver = LensEquationSolver(x_grid, y_grid, ray_shooting_func)
        return self._solver

    def image_positions_and_amplitudes(self, kwargs_point_source, 
                                       kwargs_lens=None, kwargs_solver=None,
                                       zero_amp_duplicates=True, re_compute=False):
        """Compute image plane positions and corresponding amplitudes
        of the point source, optionally "turning-off" (zeroing their amplitude)
        potentially duplicated images predicted by the lens equation solver.

        Parameters
        ----------
        kwargs_point_source : list of dict
            Keyword arguments corresponding to the point source instances.
        kwargs_lens : list of dict, optional
            Keyword arguments for the lensing mass model. Default is None.
        kwargs_solver : dict, optional
            Keyword arguments for the lens equation solver. Default is None.
        zero_amp_duplicates : bool, optional
            If True, amplitude of duplicated images are forced to be zero.
            Note that it may affect point source ordering!.
            Default is True.
        re_compute : bool, optional
            If True, re-compute (solving the lens equation) image positions,
            even for point source models of type 'IMAGE_POSITIONS'.
            Default is False.
        
        Return
        ------
        theta_x, theta_y, amp : tuple of 1D arrays
            Positions (x, y) in image plane and amplitude of the lensed images.

        """
        theta_x, theta_y = self.image_positions(
            kwargs_point_source, kwargs_lens=kwargs_lens, 
            kwargs_solver=kwargs_solver, re_compute=re_compute,
        )
        amp = self.image_amplitudes(
            theta_x, theta_y, kwargs_point_source, kwargs_lens=kwargs_lens,
        )
        if zero_amp_duplicates and self.type == 'SOURCE_POSITION':
            amp, theta_x, theta_y = self._zero_amp_duplicated_images(
                amp, theta_x, theta_y, kwargs_solver,
            )
        return theta_x, theta_y, amp

    def image_positions(self, kwargs_point_source, kwargs_lens=None, kwargs_solver=None, re_compute=False):
        """Compute image plane positions of the point source.

        Parameters
        ----------
        kwargs_point_source : list of dict
            Keyword arguments corresponding to the point source instances.
        kwargs_lens : list of dict, optional
            Keyword arguments for the lensing mass model. Default is None.
        kwargs_solver : dict, optional
            Keyword arguments for the lens equation solver. Default is None.

        """
        if self.type == 'IMAGE_POSITIONS' and not re_compute:
            theta_x = jnp.atleast_1d(kwargs_point_source['ra'])
            theta_y = jnp.atleast_1d(kwargs_point_source['dec'])
            return theta_x, theta_y
        elif self.type == 'SOURCE_POSITION' or re_compute:
            if self.type == 'IMAGE_POSITIONS':  # i.e. re_compute = True
                beta_x, beta_y = self.source_position(kwargs_point_source, kwargs_lens=kwargs_lens)
            else:
                beta_x, beta_y = kwargs_point_source['ra'], kwargs_point_source['dec']
            # Solve the lens equation
            beta = jnp.array([beta_x, beta_y])
            if kwargs_solver is None:
                kwargs_solver = {}  # fall back to default lens equation solver settings
            theta, beta = self.solver.solve(
                beta, kwargs_lens, **kwargs_solver,
            )
            return theta.T

    def image_amplitudes(self, theta_x, theta_y, kwargs_point_source, kwargs_lens=None):
        """Determine the amplitudes of the multiple images of the point source.

        Parameters
        ----------
        theta_x : array_like
            X position of points in the image plane [arcsec].
        theta_y : array_like
            Y position of points in the image plane [arcsec].
        kwargs_point_source : list of dict
            Keyword arguments corresponding to the point source instances.
        kwargs_lens : list of dict, optional
            Keyword arguments for the lensing mass model. Default is None.

        """
        amp = kwargs_point_source['amp']
        if self.type == 'IMAGE_POSITIONS':
            return jnp.atleast_1d(amp)
        elif self.type == 'SOURCE_POSITION':
            mag = self.mass_model.magnification(theta_x, theta_y, kwargs_lens)
            return amp * jnp.abs(mag)
            
    def source_position(self, kwargs_point_source, kwargs_lens=None):
        """Compute the source plane position of the point source.

        Parameters
        ----------
        kwargs_point_source : list of dict
            Keyword arguments corresponding to the point source instances.
        kwargs_lens : list of dict, optional
            Keyword arguments for the lensing mass model. Default is None.

        """
        if self.type == 'IMAGE_POSITIONS':
            theta_x = jnp.array(kwargs_point_source['ra'])
            theta_y = jnp.array(kwargs_point_source['dec'])
            beta = self.mass_model.ray_shooting(theta_x, theta_y, kwargs_lens)
            return jnp.mean(beta[0]), jnp.mean(beta[1])
        elif self.type == 'SOURCE_POSITION':
            beta_x = jnp.atleast_1d(kwargs_point_source['ra'])
            beta_y = jnp.atleast_1d(kwargs_point_source['dec'])
            return beta_x, beta_y

    def source_amplitude(self, kwargs_point_source, kwargs_lens=None):
        """Determine the amplitude of the point source in the source plane.

        Parameters
        ----------
        kwargs_point_source : list of dict
            Keyword arguments corresponding to the point source instances.
        kwargs_lens : list of dict, optional
            Keyword arguments for the lensing mass model. Default is None.

        """
        if self.type == 'IMAGE_POSITIONS':
            theta_x = jnp.atleast_1d(kwargs_point_source['ra'])
            theta_y = jnp.atleast_1d(kwargs_point_source['dec'])
            mag = self.mass_model.magnification(theta_x, theta_y, kwargs_lens)
            amps = jnp.atleast_1d(kwargs_point_source['amp']) / abs(mag)
            return jnp.mean(amps)
        elif self.type == 'SOURCE_POSITION':
            return jnp.array(kwargs_point_source['amp'])
        
    def error_image_plane(self, kwargs_point_source, kwargs_lens, kwargs_solver):
        self._check_solver_install("log_prob_image_plane")
        # get the optimized image positions
        theta_x_opti = jnp.array(kwargs_point_source['ra'])
        theta_y_opti = jnp.array(kwargs_point_source['dec'])
        # find source position via ray-tracing
        beta = self.mass_model.ray_shooting(theta_x_opti, theta_y_opti, kwargs_lens)
        beta_x, beta_y = beta[0].mean(), beta[1].mean()
        beta = jnp.array([beta_x, beta_y])
        # solve lens equation to find the predicted image positions
        theta, beta = self.solver.solve(
            beta, kwargs_lens, **kwargs_solver,
        )
        theta_x_pred, theta_y_pred = theta.T
        # return departures between original and new positions
        return jnp.sqrt((theta_x_opti - theta_x_pred)**2 + (theta_y_opti - theta_y_pred)**2)
        
    def log_prob_image_plane(self, kwargs_point_source, kwargs_lens, 
                             kwargs_solver, sigma_image=1e-3):
        error_image = self.error_image_plane(kwargs_point_source, kwargs_lens, kwargs_solver)
        # penalize departures between original and new positions
        return - jnp.sum((error_image / sigma_image)**2)
    
    def error_source_plane(self, kwargs_point_source, kwargs_lens):
        # find source position via ray-tracing
        theta_x_in = jnp.array(kwargs_point_source['ra'])
        theta_y_in = jnp.array(kwargs_point_source['dec'])
        beta_x, beta_y = self.mass_model.ray_shooting(theta_x_in, theta_y_in, kwargs_lens)
        # compute distance between mean position and ray-traced positions
        return jnp.sqrt((beta_x - beta_x.mean())**2 + (beta_y - beta_y.mean())**2)
    
    def log_prob_source_plane(self, kwargs_point_source, kwargs_lens, sigma_source=1e-3):
        error_source = self.error_source_plane(kwargs_point_source, kwargs_lens)
        return - jnp.sum((error_source / sigma_source)**2)

    def _zero_amp_duplicated_images(self, amp_in, theta_x_in, theta_y_in, kwargs_solver):
        """This function takes as input the list of multiply lensed images 
        (amplitudes and positions) and assign zero amplitude to any image 
        that have a x coordinate equal to up to `decimals` decimals.

        WARNING: this function may change the original ordering of images!

        Parameters
        ----------
        amp_in : array_like
            Amplitude of point sources
        theta_x : array_like
            X position of point sources in the image plane.
        theta_y : array_like
            Y position of point sources in the image plane.
        kwargs_solver : dict
            Keyword arguments for the LensEquation solver, used to estimate the
            accuracy of point source positions and use it to find duplicated images. 

        Returns
        -------
        amp_out, theta_x_out, theta_y_out : tuple of 3 1D arrays
            Amplitudes (potentially some being zero-ed) and positions in image plane.
        """
        # TODO: find a way not to change the image ordering (might be slower though).
        num_images = kwargs_solver['nsolutions']
        position_accuracy = self.solver.estimate_accuracy(
            kwargs_solver['niter'], 
            kwargs_solver['scale_factor'], 
            kwargs_solver['nsubdivisions'], 
        )
        # TODO: the following choice for truncation the digits may not be general enough!
        position_decimals = np.floor(- np.log10(position_accuracy)).astype(int) - 1
        unique_theta_x, unique_indices = jnp.unique(
            jnp.round(theta_x_in, decimals=position_decimals),  # TODO: issue when original value close to zero -> rounded to exactly zero!
            return_index=True, 
            fill_value=False,   # effectively zero
            size=num_images,
        )
        condition = jnp.where(unique_theta_x, True, False)
        unique_amp = amp_in[unique_indices]  # order amplitudes as the positions
        zero_amp = 1e-20  # not exactly 0 to avoid problems with autodiff gradients
        amp_out = jnp.where(condition, unique_amp, jnp.full(num_images, zero_amp))
        theta_x_out = theta_x_in[unique_indices]
        theta_y_out = theta_y_in[unique_indices]
        return amp_out, theta_x_out, theta_y_out

    def _check_solver_install(self, feature):
        if not _solver_installed:
            raise RuntimeError(f"A lens equation solver is required for the "
                               f"require point source modeling feature ('{feature}'). "
                               f"Please install `helens` from https://github.com/Herculens/helens.")
        


# Describes a light model, as a list of light profiles
# 
# Copyright (c) 2023, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LightModel module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import numpy as np
import jax.numpy as jnp

import herculens.LightModel.profile_mapping as pm
from herculens.Util import util

__all__ = ['LightModelBase']


SUPPORTED_MODELS = pm.SUPPORTED_MODELS
STRING_MAPPING = pm.STRING_MAPPING


class LightModelBase(object):
    """Base class for source and lens light models."""
    # TODO: instead of settings for creating PixelGrid objects, pass directly the object to the LightModel
    def __init__(self, profile_list, kwargs_pixelated=None,
                 **profile_specific_kwargs):
        """Create a LightModelBase object.

        NOTE: the extra keyword arguments are given to the corresponding profile class
        only when that profile is given as a string instead of a class instance.

        Parameters
        ----------
        profile_list : list of strings or profile instances
            List of mass profiles. If not a list, wrap the passed argument in a list. 
        kwargs_pixelated : dict
            Settings related to the creation of the pixelated grid.
            See herculens.PixelGrid.create_model_grid for details.
        profile_specific_kwargs : dict
            See docstring for get_class_from_string().

        """
        if not isinstance(profile_list, (list, tuple)):
            raise TypeError("The profile list should be a list or a tuple.")
        self.func_list, self._pix_idx = self._load_model_instances(
            profile_list, **profile_specific_kwargs
        )
        self._num_func = len(self.func_list)
        self._model_list = profile_list
        if kwargs_pixelated is None:
            kwargs_pixelated = {}
        self._kwargs_pixelated = kwargs_pixelated
        
    def _load_model_instances(
            self, profile_list, **profile_specific_kwargs,
        ):
        func_list = []
        pix_idx = None
        for idx, profile_type in enumerate(profile_list):
            if isinstance(profile_type, str):
                # passing string is supported for backward-compatibility only
                profile_class = self.get_class_from_string(
                    profile_type, 
                    **profile_specific_kwargs,
                )
                if profile_type in ['PIXELATED']:
                    pix_idx = idx

            # this is the new preferred way: passing the profile as a class
            elif self.is_light_profile_class(profile_type):
                profile_class = profile_type
                if isinstance(profile_class, STRING_MAPPING['PIXELATED']):
                    pix_idx = idx
            else:
                raise ValueError("Each profile can either be a string or "
                                 "directly the profile instance.")
            func_list.append(profile_class)
        return func_list, pix_idx
    
    @staticmethod
    def is_light_profile_class(profile):
        """Simply checks that the mass profile has the required methods"""
        return hasattr(profile, 'function')

    @staticmethod
    def get_class_from_string(
            profile_string, 
            smoothing=0.001, 
            shapelets_n_max=4,
            superellipse_exponent=2, 
            pixel_interpol='bilinear', 
            pixel_adaptive_grid=False, 
            pixel_allow_extrapolation=False,
        ):
        """
        Get the profile class of the corresponding type.
        Keyword arguments are related to specific profile types.
            
        Parameters
        ----------
        smoothing : float
            Smoothing factor for some models.
        shapelets_n_max : int
            Maximal order of the shapelets basis set.
        superellipse_exponent : int, float
            Exponent for super-elliptical profiles (e.g. 'SERSIC_SUPERELLIPSE').
        pixel_interpol : string
            Type of interpolation for 'PIXELATED' profiles: 'bilinear' or 'bicubic'
        pixel_adaptive_grid : bool
            Whether or not the pixelated light profile is defined on a grid
            whose extent is adapted based on other model components.
        pixel_allow_extrapolation : bool
            Wether or not to allow the interpolator to predict values outside 
            the field of view of the pixelated profile
        """
        if profile_string in SUPPORTED_MODELS:
            profile_class = STRING_MAPPING[profile_string]
            # treats the few special cases that require user settings
            if profile_string == 'SERSIC':
                return profile_class(smoothing=smoothing)
            elif profile_string == 'SERSIC_ELLIPSE':
                return profile_class(smoothing=smoothing, exponent=2)
            elif profile_string == 'SERSIC_SUPERELLIPSE':
                return profile_class(smoothing=smoothing, exponent=superellipse_exponent)
            elif profile_string == 'SHAPELETS':
                return profile_class(shapelets_n_max)
            elif profile_string == 'PIXELATED':
                return profile_class(interpolation_type=pixel_interpol, 
                                    allow_extrapolation=pixel_allow_extrapolation, 
                                    adaptive_grid=pixel_adaptive_grid)
        else:
            raise ValueError(f"Could not load profile type '{profile_string}'.")
        # all remaining profiles take no extra arguments
        return profile_class()

    @property
    def param_name_list(self):
        """Get parameter names as a list of strings for each light model."""
        return [func.param_names for func in self.func_list]

    def _bool_list(self, k):
        return util.convert_bool_list(n=self._num_func, k=k)

    @property
    def has_pixels(self):
        return self._pix_idx is not None

    @property
    def pixel_grid_settings(self):
        return self._kwargs_pixelated

    def set_pixel_grid(self, pixel_grid, data_pixel_area):
        self.func_list[self.pixelated_index].set_pixel_grid(pixel_grid, data_pixel_area)

    @property
    def pixel_grid(self):
        if not self.has_pixels:
            return None
        return self.func_list[self.pixelated_index].pixel_grid

    @property
    def pixelated_index(self):
        # TODO: support multiple pixelated profiles
        return self._pix_idx

    @property
    def pixelated_coordinates(self):
        if not self.has_pixels:
            return None, None
        return self.pixel_grid.pixel_coordinates

    @property
    def pixelated_shape(self):
        if not self.has_pixels:
            return None
        x_coords, _ = self.pixelated_coordinates
        return x_coords.shape
    
    @property
    def pixel_is_adaptive(self):
        if not self.has_pixels:
            return False
        return self.func_list[self.pixelated_index].is_adaptive

    @property
    def num_amplitudes_list(self):
        return [func.num_amplitudes for func in self.func_list]




# High-level interface to a light model
# 
# Copyright (c) 2023, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the LightModel module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'

from functools import partial
import jax.numpy as jnp

from herculens.LightModel.light_model_base import LightModelBase


__all__ = ['LightModel']


def function_static_single(
        x, y, function,
    ):
    return partial(function, x, y)


class LightModel(LightModelBase):
    """Model extended surface brightness profiles of sources and lenses.

    Notes
    -----
    All profiles come with a surface_brightness parameterization (in units per
    square angle and independent of the pixel scale.) The parameter `amp` is
    the linear scaling parameter of surface brightness. Some profiles have
    a total_flux() method that gives the integral of the surface brightness
    for a given set of parameters.

    """
    def __init__(self, profile_list, verbose=False, **kwargs):
        """Create a LightModel object.

        Parameters
        ----------
        profile_list : list of strings or profile instances
            List of light profiles.
        kwargs_pixelated : dictionary for settings related to PIXELATED profiles.
        """
        if not isinstance(profile_list, (list, tuple)):
            # useful when using a single profile
            profile_list = [profile_list]
        self.profile_type_list = profile_list
        super(LightModel, self).__init__(self.profile_type_list, **kwargs)
        self._repeated_profile_mode = False
        self._single_profile_mode = len(self.profile_type_list) == 1
        if len(self.profile_type_list) > 0:
            first_profile = self.profile_type_list[0]
            self._repeated_profile_mode = (
                all(p is first_profile for p in self.profile_type_list)
            )
            if verbose is True and self._repeated_profile_mode:
                print("All LightModel profiles are identical.")

    def surface_brightness(self, x, y, kwargs, k=None,
                           pixels_x_coord=None, pixels_y_coord=None,
                           return_as_list=False):
            """Total source flux at a given position.

            Parameters
            ----------
            x : float or array_like
                Position coordinate(s) in arcsec relative to the image center.
            y : float or array_like
                Position coordinate(s) in arcsec relative to the image center.
            kwargs : list
                List of parameter dictionaries corresponding to each source model.
            k : int, optional
                Position index of a single source model component.
            pixels_x_coord : array_like, optional
                x-coordinates of the pixelated light profile (if any).
            pixels_y_coord : array_like, optional
                y-coordinates of the pixelated light profile (if any).
            return_as_list : bool, optional
                If True, return the flux of each profile separately.

            Returns
            -------
            float or array_like
                Total source flux at the given position(s).

            """
            # x = np.array(x, dtype=float)
            # y = np.array(y, dtype=float)
            if isinstance(k, int):
                return self._surf_bright_single(x, y, kwargs, k=k,
                                                pixels_x_coord=pixels_x_coord,
                                                pixels_y_coord=pixels_y_coord,
                                                return_as_list=return_as_list)
            elif self._single_profile_mode:
                return self._surf_bright_single(x, y, kwargs, k=0,
                                                pixels_x_coord=pixels_x_coord,
                                                pixels_y_coord=pixels_y_coord,
                                                return_as_list=return_as_list)
            elif self._repeated_profile_mode:
                return self._surf_bright_repeated(x, y, kwargs, k=k,
                                                  pixels_x_coord=pixels_x_coord,
                                                  pixels_y_coord=pixels_y_coord,
                                                  return_as_list=return_as_list)
            else:
                return self._surf_bright_loop(x, y, kwargs, k=k,
                                              pixels_x_coord=pixels_x_coord,
                                              pixels_y_coord=pixels_y_coord,
                                              return_as_list=return_as_list)
            
    def _surf_bright_single(self, x, y, kwargs, k=None,
                            pixels_x_coord=None, pixels_y_coord=None,
                            return_as_list=False):
        if k == self.pixelated_index:
            flux = self.func_list[k].function(
                x, y, **kwargs[k],
                pixels_x_coord=pixels_x_coord, 
                pixels_y_coord=pixels_y_coord,
            )
        else:
            flux = self.func_list[k].function(x, y, **kwargs[k])
        if return_as_list:
            return [flux]
        return flux
        
    def _surf_bright_repeated(self, x, y, kwargs, k=None,
                              pixels_x_coord=None, pixels_y_coord=None,
                              return_as_list=False):
        if k is not None:
            raise NotImplementedError("Repeated profile mode not implemented "
                                      "specific profile k.")
        func = function_static_single(x, y, self.func_list[0].function)
        flux_list = [
            func(**kwargs[i]) for i in range(self._num_func)
        ]
        if return_as_list:
            return flux_list
        return jnp.sum(jnp.array(flux_list),axis=0)

    def _surf_bright_loop(self, x, y, kwargs_list, k=None,
                          pixels_x_coord=None, pixels_y_coord=None,
                          return_as_list=False):
        if return_as_list:
            flux = []
        else:
            flux = jnp.zeros_like(x)
        bool_list = self._bool_list(k)
        for i, func in enumerate(self.func_list):
            if bool_list[i]:
                if i == self.pixelated_index:
                    flux_i = func.function(x, y, 
                                           pixels_x_coord=pixels_x_coord, 
                                           pixels_y_coord=pixels_y_coord, 
                                           **kwargs_list[i])
                else:
                    flux_i = func.function(x, y, **kwargs_list[i])
                if return_as_list:
                    flux.append(flux_i)
                else:
                    flux += flux_i
        return flux

    def spatial_derivatives(self, x, y, kwargs_list, k=None):
        """Spatial derivatives of the source flux at a given position (along x and y directions).

        Parameters
        ----------
        x, y : float or array_like
            Position coordinate(s) in arcsec relative to the image center.
        kwargs_list : list
            List of parameter dictionaries corresponding to each source model.
        k : int, optional
            Position index of a single source model component.

        """
        # x = jnp.array(x, dtype=float)
        # y = jnp.array(y, dtype=float)
        f_x, f_y = jnp.zeros_like(x), jnp.zeros_like(x)
        bool_list = self._bool_list(k)
        for i, func in enumerate(self.func_list):
            if bool_list[i]:
                f_x_, f_y_ = func.derivatives(x, y, **kwargs_list[i])
                f_x += f_x_
                f_y += f_y_
        return f_x, f_y



# Utility class to load GLEE files and convert to Herculens syntax

__author__ = 'aymgal'

import numpy as np
from pprint import pprint


class GLEEReader(object):
    """Load and parse a GLEE config file.

    NOTE: As of now, not all the config file can be parsed.
    Things that are not parsed are, for instance:
    - a:X,X,X statements
    - step:X,X statements
    - min:X and max:X
    - a large part of the esources properties
    - cosmology

    Parameters
    ----------
    config_file_path : str
        Path to the config file.
    verbose : bool
        Whether or not to print warning messages. Default to False.
    """

    _num_param_for_profile = {
        # TODO: add other mass/light profiles
        'dpie': 7,
        # 'piemd': 7,
        'shear': 2,
    }
    _profile_names = list(_num_param_for_profile.keys())

    def __init__(self, config_file_path, verbose=False):
        self._config_path = config_file_path
        self._verbose = verbose
        self._parsed = False

    def print_summary(self):
        if not self._parsed:
            self._raise_parser_run_error()
        print("="*60)
        print("Parsed GLEE model:")
        print("-"*18)
        print("> Number of lenses:", self.num_lenses)
        print("> Number of point-like sources:", self.num_point_like_sources)
        print(f"  leading to {self.num_point_like_images} point-like multiple images")
        print("> Number of extended sources:", self.num_extended_sources)
        print("> Number of unique source redshifts:", self.num_source_planes)
        print("="*60)

    @property
    def num_lenses(self):
        return len(self.lens_parameters)

    @property
    def num_point_like_sources(self):
        return len(self.point_like_source_parameters)

    @property
    def num_point_like_images(self):
        params = self.point_like_source_parameters
        return sum([len(p['x_img']) for p in params])

    @property
    def num_extended_sources(self):
        return len(self.extended_source_parameters)
        
    @property
    def lens_redshifts(self):
        redshifts = [p['z'] for p in self.lens_parameters]
        redshifts_priors = [p['z'] for p in self.lens_priors]
        return redshifts, redshifts_priors
    
    @property
    def lens_profiles(self):
        if not hasattr(self, '_lens_profiles'):
            self._raise_parser_run_error()
        return self._lens_profiles
    
    @property
    def lens_parameters(self):
        if not hasattr(self, '_lens_params'):
            self._raise_parser_run_error()
        return self._lens_params

    @property
    def lens_priors(self):
        if not hasattr(self, '_lens_priors'):
            self._raise_parser_run_error()
        return self._lens_priors

    @property
    def lens_labels(self):
        if not hasattr(self, '_lens_labels'):
            self._raise_parser_run_error()
        return self._lens_labels
    
    @property
    def lens_scalings(self):
        if not hasattr(self, '_lens_scalings'):
            self._raise_parser_run_error()
        return self._lens_scalings

    @property
    def point_like_source_redshifts(self):
        redshifts = [p['z'] for p in self.point_like_source_parameters]
        redshifts_priors = [p['z'] for p in self.point_like_source_priors]
        return redshifts, redshifts_priors
    
    @property
    def point_like_source_parameters(self):
        if not hasattr(self, '_ptl_src_params'):
            self._raise_parser_run_error()
        return self._ptl_src_params

    @property
    def point_like_source_priors(self):
        if not hasattr(self, '_ptl_src_priors'):
            self._raise_parser_run_error()
        return self._ptl_src_priors

    @property
    def point_like_source_labels(self):
        if not hasattr(self, '_ptl_src_labels'):
            self._raise_parser_run_error()
        return self._ptl_src_labels

    @property
    def point_like_source_errors(self):
        if not hasattr(self, '_ptl_src_errors'):
            self._raise_parser_run_error()
        return self._ptl_src_errors
    
    @property
    def extended_source_redshifts(self):
        redshifts = [p['z'] for p in self.extended_source_parameters]
        redshifts_priors = [p['z'] for p in self.extended_source_priors]
        return redshifts, redshifts_priors
    
    @property
    def extended_source_parameters(self):
        if not hasattr(self, '_ext_src_params'):
            self._raise_parser_run_error()
        return self._ext_src_params

    @property
    def extended_source_priors(self):
        if not hasattr(self, '_ext_src_priors'):
            self._raise_parser_run_error()
        return self._ext_src_priors

    @property
    def extended_source_settings(self):
        if not hasattr(self, '_ext_src_settings'):
            self._raise_parser_run_error()
        return self._ext_src_settings

    @property
    def extended_source_ref_coordinates(self):
        return [s['refcoord'] for s in self.extended_source_settings]
    
    def source_plane_groups(self):
        """Return a a list of unique redshifts of all planes source containing 'sources', as well as
        the list of lists containing indices of point-like sources
        which have the redshift, order by increasing redshift.
        For instance: [
            [redshift1_index1, redshift1_index2],  # redshift 1
            [redshift2_index1],  # redshift 2
            [redshift3_index1, redshift3_index2, redshift3_index3],  # redshift 3
            ...
        ]
        and a list of booleans indicating if the redshift plane contains 
        an extended source or not.

        Returns
        -------
        list
            List lists of integers
        """
        redshifts = np.array(self.point_like_source_redshifts[0])
        redshifts_extended = np.array(self.extended_source_redshifts[0])
        common_redshifts = np.isin(redshifts_extended, redshifts).all()
        if not common_redshifts:
            raise NotImplementedError("Extended sources can only have redshifts that " \
            "are also in point-like sources, for now.")
        iter_indices = [int(i) for i in np.argsort(redshifts)]
        group_indices = [
            [iter_indices[0]]
        ]
        unique_redshifts = [redshifts[iter_indices[0]]]
        extended_flags = [True] if unique_redshifts[0] in redshifts_extended else [False]
        for i in iter_indices[1:]:
            z_current = redshifts[i]
            z_prev = redshifts[i-1]
            if z_current != z_prev:
                group_indices.append([])
                unique_redshifts.append(z_current)
                extended_flags.append(True if z_current in redshifts_extended else False)
            group_indices[-1].append(i)
        return np.array(unique_redshifts), group_indices, extended_flags
    
    @property
    def num_source_planes(self):
        """Return the number of unique source planes.

        Returns
        -------
        int
            Number of source planes
        """
        redshifts, _, _ = self.source_plane_groups()
        return len(redshifts)

    def parse_config(self):
        lines = self.read_and_separate_lines(self._config_path)
        # next we go through the lines and group them by lens/source/esource blocks
        model_component_blocks = self._extract_blocks(
            lines, ['lenses_vary', 'sources', 'esources'],
        )
        # parse the lens block
        self._lens_profiles, self._lens_params, self._lens_priors, self._lens_labels, self._lens_scalings \
            = self._parse_lens_model(model_component_blocks['lenses_vary'])
        # parse the point-like source block
        self._ptl_src_params, self._ptl_src_priors, self._ptl_src_labels, self._ptl_src_errors \
            = self._parse_point_like_source_model(model_component_blocks['sources'])
        # parse the extended source block
        self._ext_src_params, self._ext_src_priors, self._ext_src_settings \
            = self._parse_extended_sources_block(model_component_blocks['esources'])
        # if successful so far, then we say it's parsed
        self._parsed = True

    def _parse_lens_model(self, block):
        header = block[0]
        content = block[1:]
        component_type = header[0]
        num_components = int(header[1])
        model_profile_blocks, profile_types = self._extract_blocks(
            content, self._profile_names, with_duplicates=True,
        )
        assert num_components == len(profile_types), f"Inconsistent number of lens profiles ('{component_type}')"
        return self._parse_all_lens_profiles(profile_types, model_profile_blocks)

    def _parse_all_lens_profiles(self, types, blocks):
        params_list = []
        priors_list = []
        labels_list = []
        scalings_list = []
        for i, (type_, block) in enumerate(zip(types, blocks)):
            header = block[0]
            content = block[1:]
            assert self.get_line_item(header) == type_, f"Inconsistent lens profile type ({i})"
            kwargs_params, kwargs_priors, kwargs_labels, kwargs_scalings = self._parse_lens_profile(i, content)
            params_list.append(kwargs_params)
            priors_list.append(kwargs_priors)
            labels_list.append(kwargs_labels)
            scalings_list.append(kwargs_scalings)
        return types, params_list, priors_list, labels_list, scalings_list
    
    def _parse_lens_profile(self, i, lines):
        redshift, redshift_prior, redshift_label = self._parse_redshift(i, lines[0])
        kwargs_params, kwargs_priors, kwargs_labels, kwargs_scalings = self._parse_lens_parameters(lines[1:])
        # add the redshift to the rest of the parameters
        kwargs_params['z'] = redshift
        kwargs_priors['z'] = redshift_prior
        kwargs_labels['z'] = redshift_label
        return kwargs_params, kwargs_priors, kwargs_labels, kwargs_scalings

    def _parse_lens_parameters(self, lines):
        kwargs_params = {}
        kwargs_priors = {}
        kwargs_labels = {}
        kwargs_scalings = {}
        for line in lines:
            if self.line_is_empty(line):
                continue
            value = float(self.get_line_item(line, pos=0))
            name = self.get_line_item(line, pos=1).replace('#', '')
            prior = self._parse_prior(line)
            label = self._parse_label(line)
            scaling = self._parse_scaling(line)
            kwargs_params[name] = value
            kwargs_priors[name] = prior
            kwargs_labels[name] = label
            kwargs_scalings[name] = scaling
        return kwargs_params, kwargs_priors, kwargs_labels, kwargs_scalings

    def _parse_point_like_source_model(self, block):
        header = block[0]
        content = block[1:]
        component_type = header[0]
        num_components = int(header[1])
        model_component_blocks, _ = self._extract_blocks(
            content, ['z'], with_duplicates=True,  # we extract the block that starts with z
        )
        assert (num_components == len(model_component_blocks), 
                f"Inconsistent number of point-like sources ('{component_type}')")
        return self._parse_all_point_like_sources(model_component_blocks)

    def _parse_all_point_like_sources(self, blocks):
        params_list = []
        priors_list = []
        labels_list = []
        errors_list = []
        for i, block in enumerate(blocks):
            kwargs_params, kwargs_priors, kwargs_errors, kwargs_labels \
                = self._parse_point_like_source(i, block)
            params_list.append(kwargs_params)
            priors_list.append(kwargs_priors)
            labels_list.append(kwargs_labels)
            errors_list.append(kwargs_errors)
        return params_list, priors_list, labels_list, errors_list

    def _parse_point_like_source(self, i, lines):
        redshift, redshift_prior, redshift_label = self._parse_redshift(i, lines[0])
        kwargs_params, kwargs_priors, kwargs_errors = self._parse_ptl_src_parameters(i, lines[1:])
        # add the redshift to the rest of the parameters
        kwargs_params['z'] = redshift
        kwargs_priors['z'] = redshift_prior
        kwargs_errors['z'] = None
        kwargs_labels = {'z': redshift_label}
        return kwargs_params, kwargs_priors, kwargs_errors, kwargs_labels

    def _parse_ptl_src_parameters(self, i, lines):
        # type of point-like source weighting (?)
        assert self.get_line_item(lines[0], pos=0) == 'source'  # sanity check
        src_type = self.get_line_item(lines[0], pos=1)
        # source plane position coordinates
        assert self.get_line_item(lines[1], pos=0) == 'srcx'  # sanity check
        x_src = float(self.get_line_item(lines[1], pos=1))
        prior_x_src = self._parse_prior(lines[1])
        assert self.get_line_item(lines[2], pos=0) == 'srcy'  # sanity check
        y_src = float(self.get_line_item(lines[2], pos=1))
        prior_y_src = self._parse_prior(lines[2])
        # number of multiple images
        num_images = int(self.get_line_item(lines[3]))
        # image plane positions coordinates
        x_img, y_img = [], []
        err_img = []
        for line in lines[4:]:
            if self.line_is_empty(line):
                continue
            x_img.append(float(self.get_line_item(line, pos=0)))
            y_img.append(float(self.get_line_item(line, pos=1)))
            err_img.append(self._parse_error(line))
        assert len(x_img) == len(y_img) == num_images, f"Inconsistent number of multiple images for point-like source {i}."
        kwargs_params = {
            'x_src': x_src,
            'y_src': y_src,
            'x_img': x_img,
            'y_img': y_img,
        }
        kwargs_priors = {
            'x_src': prior_x_src,
            'y_src': prior_y_src,
        }
        kwargs_errors = {
            'x_img': err_img,
            'y_img': err_img,
        }
        return kwargs_params, kwargs_priors, kwargs_errors
    
    def _parse_extended_sources_block(self, block):
        header = block[0]
        content = block[1:]
        component_type = header[0]
        num_components = int(header[1])
        model_component_blocks, _ = self._extract_blocks(
            content, ['z'], with_duplicates=True,  # we extract the block that starts with z
        )
        assert num_components == len(model_component_blocks), f"Inconsistent number of extended sources ('{component_type}')"
        return self._parse_all_extended_sources(model_component_blocks)

    def _parse_all_extended_sources(self, blocks):
        params_list = []
        priors_list = []
        settings_list = []
        for i, block in enumerate(blocks):
            kwargs_params, kwargs_priors, kwargs_settings \
                = self._parse_extended_source(i, block)
            params_list.append(kwargs_params)
            priors_list.append(kwargs_priors)
            settings_list.append(kwargs_settings)
        return params_list, priors_list, settings_list

    def _parse_extended_source(self, i, lines):
        redshift, redshift_prior, _ = self._parse_redshift(i, lines[0])
        kwargs_settings = self._parse_extended_source_settings(i, lines[1:])
        # add the redshift to the rest of the parameters
        kwargs_params = {'z': redshift}
        kwargs_priors = {'z': redshift_prior}
        return kwargs_params, kwargs_priors, kwargs_settings

    def _parse_extended_source_settings(self, i, lines):
        kwargs_settings = {}
        for line in lines:
            setting_name = self.get_line_item(line, pos=0)
            if setting_name is None or setting_name == 'esource_end':
                break
            elif setting_name == 'esource_refcoord':
                ref_x = float(self.get_line_item(line, pos=1))
                ref_y = float(self.get_line_item(line, pos=2))
                setting_value = (ref_x, ref_y)
            else:
                setting_value = self.get_line_item(line, pos=1)
            setting_name = setting_name.replace('esource_', '')  # remove the 'esource' prefix
            kwargs_settings[setting_name] = setting_value
        return kwargs_settings
    
    def _parse_redshift(self, i, line):
        assert self.get_line_item(line) == 'z'
        z = self.get_line_item(line, pos=1)  # redshift value is the next one
        prior = self._parse_prior(line)
        label = self._parse_label(line)
        return float(z), prior, label

    def _parse_prior(self, line):
        # TODO: replace pos=2 with a list of start_with
        prior_item = self.get_line_item(line, pos=2).split(':')
        prior_type = prior_item[0]
        if prior_type in ('exact', 'noprior'):
            val1, val2 = None, None
        elif prior_type == 'flat':
            bounds = prior_item[1].split(',')
            val1 = float(bounds[0])
            val2 = float(bounds[1])
        elif prior_type == 'link':
            link_to_label = prior_item[1]
            val1 = link_to_label
            val2 = None
        else:
            raise ValueError(f"Prior type '{prior_type}' is not supported")
        return (prior_type, val1, val2)

    def _parse_error(self, line):
        error_item = self.get_line_item(line, start_with='error:')
        if error_item is None:
            return None
        error_item = error_item.split(':')
        error_type = error_item[0]
        error_value = float(error_item[1])
        assert error_type == 'error'  # sanity check
        return error_value

    def _parse_label(self, line):
        label_item = self.get_line_item(line, start_with='label:')
        if label_item is None:
            return None
        label = label_item.split(':')[1]
        return label
    
    def _parse_scaling(self, line):
        scaling_item = self.get_line_item(line, start_with='a:')
        if scaling_item is None:
            return (None, None, None)
        scaling_item = scaling_item.split(':')[1].split(',')
        coeff_add = float(scaling_item[0])
        coeff_mul = float(scaling_item[1])
        coeff_pow = float(scaling_item[2])
        return (coeff_add, coeff_mul, coeff_pow)
    
    def _extract_blocks(self, lines, separator_strings, with_duplicates=False):
        num_lines = len(lines)
        lines.append([[None]])  # just for practicality, adds a line with just None
        if with_duplicates:
            # if there are several blocks with the same separator, we use two lists
            # to allow for duplicates
            block_keys = []
            blocks = []
        else:
            # in this case we use a dictionary as we group all blocks
            # that belong to the same separator
            blocks = {k: [] for k in separator_strings}
        current_block_type = None
        block_is_done = True
        for i in range(num_lines):
            line_items = lines[i]
            first_item = self.get_line_item(line_items)
            next_first_item = self.get_line_item(lines[i+1])
            if first_item in separator_strings:
                current_block_type = first_item
                block_is_done = False
                if with_duplicates:
                    block_keys.append(current_block_type)
                    blocks.append([])
            if not block_is_done:
                if with_duplicates:
                    blocks[-1].append(line_items)
                else:
                    blocks[current_block_type].append(line_items)
            if next_first_item in separator_strings:
                block_is_done = True
        if with_duplicates:
            return blocks, block_keys
        return blocks

    @staticmethod
    def get_line_item(line, pos=0, start_with=None):
        """
        Given a 'line' defined as a list of strings, get the next (if pos=0) 
        non-empty (i.e. different from ''), or any one after if pos>0.
        """
        if start_with is not None and not isinstance(start_with, (str, tuple, list)):
            raise ValueError("`start_with` can only be None, a string or a list/tuple of strings.")
        i = 0
        for item in line:
            if start_with is None:
                if item != '':
                    if pos == i:
                        return item
                    i += 1
            elif isinstance(start_with, str):
                if start_with in item:
                    return item
            else:
                for start_with_sgl in start_with:
                    if start_with_sgl in item:
                        return item
        return None  # if nothing was found

    @staticmethod
    def line_is_empty(line):
        return len(line) == 0 or all([l == '' for l in line])

    @staticmethod
    def read_and_separate_lines(file_path):
        with open(file_path) as f:
            lines = f.readlines()
        # remove \n and split each line into individual items
        lines = [l.replace('\n', '').split(' ') for l in lines]
        return lines

    def _print(self, txt):
        if self._verbose is True:
            print(txt)
    
    def _raise_parser_run_error(self):
        raise RuntimeError("You must run parse_config() beforehand.")



# Utility functions
#
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the Util module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import numpy as np
import jax.numpy as jnp
import json
import re


def rotate(xcoords, ycoords, angle):
    """Rotate points about the origin by an angle."""
    cos = jnp.cos(angle)
    sin = jnp.sin(angle)
    new_x =   xcoords * cos + ycoords * sin
    new_y = - xcoords * sin + ycoords * cos
    return new_x, new_y

def map_coord2pix(ra, dec, x_0, y_0, M):
    """Perform a linear transformation between two coordinate systems.

    Mainly used to transform angular into pixel coordinates in an image.
    :param ra: ra coordinates
    :param dec: dec coordinates
    :param x_0: pixel value in x-axis of ra,dec = 0,0
    :param y_0: pixel value in y-axis of ra,dec = 0,0
    :param M: 2x2 matrix to transform angular to pixel coordinates
    :return: transformed coordinate systems of input ra and dec
    """
    #x, y = M.dot(np.array([ra, dec]))
    x, y = jnp.array(M).dot(jnp.array([ra, dec]))
    return x + x_0, y + y_0

def array2image(array, nx=0, ny=0):
    """Convert a 1d array into a 2d array.

    Note: this only works when length of array is a perfect square, or else if
    nx and ny are provided

    :param array: 1d array of image values
    :type array: array of size n**2
    :returns:  2d array
    :raises: ValueError
    """
    if nx == 0 or ny == 0:
        # Avoid turning n into a JAX-traced object with jax.numpy.sqrt
        n = int(len(array)**0.5)
        if n**2 != len(array):
            err_msg = f"Input array size {len(array)} is not a perfect square."
            raise ValueError(err_msg)
        nx, ny = n, n
    image = array.reshape(int(nx), int(ny))
    return image

def image2array(image):
    """Convert a 2d array into a 1d array.

    :param array: image values
    :type array: array of size (n,n)
    :returns:  1d array
    :raises: AttributeError, KeyError
    """
    # nx, ny = image.shape  # find the size of the array
    # imgh = np.reshape(image, nx * ny)  # change the shape to be 1d
    # return imgh
    return image.ravel()

def make_grid(numPix, deltapix, subgrid_res=1, left_lower=False):
    """
    creates pixel grid (in 1d arrays of x- and y- positions)
    default coordinate frame is such that (0,0) is in the center of the coordinate grid

    :param numPix: number of pixels per axis
        Give an integers for a square grid, or a 2-length sequence
        (first, second axis length) for a non-square grid.
    :param deltapix: pixel size
    :param subgrid_res: sub-pixel resolution (default=1)
    :return: x, y position information in two 1d arrays
    """

    # Check numPix is an integer, or 2-sequence of integers
    if isinstance(numPix, (tuple, list, np.ndarray)):
        assert len(numPix) == 2
        if any(x != round(x) for x in numPix):
            raise ValueError("numPix contains non-integers: %s" % numPix)
        # numPix = np.asarray(numPix, dtype=int)
        numPix = list(numPix)
    else:
        if numPix != round(numPix):
            raise ValueError("Attempt to specify non-int numPix: %s" % numPix)
        # numPix = np.array([numPix, numPix], dtype=int)
        numPix = [numPix, numPix]

    # Super-resolution sampling
    # numPix_eff = int(numPix * subgrid_res)
    numPix_eff = [int(n * subgrid_res) for n in numPix]
    deltapix_eff = deltapix / float(subgrid_res)

    # Compute unshifted grids.
    # X values change quickly, Y values are repeated many times
    # NOTE jax.numpy.tile checks if `reps` is of type int, but numpy.int64
    #      is not in fact this type. Simply casting as int(numPix_eff[1])
    #      causes problems elsewhere with jax tracing, so we use another approach
    # x_grid = np.tile(np.arange(numPix_eff[0]), numPix_eff[1]) * deltapix_eff
    # y_grid = np.repeat(np.arange(numPix_eff[1]), numPix_eff[0]) * deltapix_eff
    x_space = np.arange(numPix_eff[0]) * deltapix_eff
    y_space = np.arange(numPix_eff[1]) * deltapix_eff
    x_grid, y_grid = np.meshgrid(x_space, y_space)
    x_grid, y_grid = x_grid.flatten(), y_grid.flatten()

    if left_lower is True:
        # Shift so (0, 0) is in the "lower left"
        # Note this does not shift when subgrid_res = 1
        shift = -1. / 2 + 1. / (2 * subgrid_res) * np.array([1, 1])
    else:
        # Shift so (0, 0) is centered
        # shift = deltapix_eff * (numPix_eff - 1) / 2
        shift = np.array([deltapix_eff * (n - 1) / 2 for n in numPix_eff])

    return x_grid - shift[0], y_grid - shift[1]

def make_grid_transformed(numPix, Mpix2Angle):
    """
    returns grid with linear transformation (deltaPix and rotation)
    :param numPix: number of Pixels
    :param Mpix2Angle: 2-by-2 matrix to mat a pixel to a coordinate
    :return: coordinate grid
    """
    x_grid, y_grid = make_grid(numPix, deltapix=1)
    ra_grid, dec_grid = map_coord2pix(x_grid, y_grid, 0, 0, Mpix2Angle)
    return ra_grid, dec_grid

def grid_from_coordinate_transform(nx, ny, Mpix2coord, ra_at_xy_0, dec_at_xy_0):
    """Return x and y coordinate grids that satisfy the coordinate system.

    :param nx: number of pixels in x-axis
    :param ny: number of pixels in y-axis
    :param Mpix2coord: transformation matrix (2x2) of pixels into coordinate displacements
    :param ra_at_xy_0: RA coordinate at (x,y) = (0,0)
    :param dec_at_xy_0: DEC coordinate at (x,y) = (0,0)
    :return: RA coordinate grid, DEC coordinate grid
    """
    a = np.arange(nx)
    b = np.arange(ny)
    matrix = np.dstack(np.meshgrid(a, b)).reshape(-1, 2)
    x_grid = matrix[:, 0]
    y_grid = matrix[:, 1]
    ra_grid = x_grid * Mpix2coord[0, 0] + y_grid * Mpix2coord[0, 1] + ra_at_xy_0
    dec_grid = x_grid * Mpix2coord[1, 0] + y_grid * Mpix2coord[1, 1] + dec_at_xy_0
    return ra_grid, dec_grid

def subgrid_from_coordinate_transform(nx, ny, Mpix2coord, ra_at_xy_0, dec_at_xy_0, subgrid_res=2):
    """

    :param nx: number of pixels in x-axis
    :param ny: number of pixels in y-axis
    :param Mpix2coord: transformation matrix (2x2) of pixels into coordinate displacements
    :param ra_at_xy_0: RA coordinate at (x,y) = (0,0)
    :param dec_at_xy_0: DEC coordinate at (x,y) = (0,0)
    :return: RA coordinate grid, DEC coordinate grid
    """
    if subgrid_res == 1:
        return grid_from_coordinate_transform(nx, ny, Mpix2coord, ra_at_xy_0, dec_at_xy_0)

    nx_sub, ny_sub = int(nx * subgrid_res), int(ny * subgrid_res)
    subgrid_res = float(subgrid_res)
    Mcoord2pix = np.linalg.inv(Mpix2coord)
    x_at_radec_0, y_at_radec_0 = map_coord2pix(ra_at_xy_0, dec_at_xy_0, x_0=0, y_0=0, M=Mcoord2pix)
    Mpix2coord_sub = Mpix2coord / subgrid_res
    x_at_radec_0_sub = x_at_radec_0 * subgrid_res - 0.5*(subgrid_res-1)
    y_at_radec_0_sub = y_at_radec_0 * subgrid_res - 0.5*(subgrid_res-1)
    ra_at_xy_0_sub, dec_at_xy_0_sub = map_coord2pix(x_at_radec_0_sub, y_at_radec_0_sub, x_0=0, y_0=0, M=Mpix2coord_sub)
    return grid_from_coordinate_transform(nx_sub, ny_sub, Mpix2coord_sub, ra_at_xy_0_sub, dec_at_xy_0_sub)

def averaging(grid, numGrid, numPix):
    """
    resize 2d pixel grid with numGrid to numPix and averages over the pixels
    :param grid: higher resolution pixel grid
    :param numGrid: number of pixels per axis in the high resolution input image
    :param numPix: lower number of pixels per axis in the output image (numGrid/numPix is integer number)
    :return:
    """

    Nbig = numGrid
    Nsmall = numPix
    small = grid.reshape([int(Nsmall), int(Nbig / Nsmall), int(Nsmall), int(Nbig / Nsmall)]).mean(3).mean(1)
    return small

def fwhm2sigma(fwhm):
    """

    :param fwhm: full-widt-half-max value
    :return: gaussian sigma (sqrt(var))
    """
    sigma = fwhm / (2 * np.sqrt(2 * np.log(2)))
    return sigma

def sigma2fwhm(sigma):
    """

    :param sigma:
    :return:
    """
    fwhm = sigma * (2 * np.sqrt(2 * np.log(2)))
    return fwhm

def make_subgrid(ra_coord, dec_coord, subgrid_res=2):
    """
    return a grid with subgrid resolution
    :param ra_coord:
    :param dec_coord:
    :param subgrid_res:
    :return:
    """
    # With JAX we need to avoid array item assignment through indexing

    ra = array2image(ra_coord)[0]  # First row containing all ra values
    n_ra, ra_min, ra_max = len(ra), ra[0], ra[-1]
    d_ra = (np.diff(ra)[0] / 2.) * (1. - 1. / subgrid_res)  # Desired spacing
    ra_new = np.linspace(ra_min - d_ra, ra_max + d_ra, subgrid_res * n_ra)

    dec = array2image(dec_coord)[:, 0]  # First column containing all dec values
    n_dec, dec_min, dec_max = len(dec), dec[0], dec[-1]
    d_dec = (np.diff(dec)[0] / 2.) * (1. - 1. / subgrid_res)
    dec_new = np.linspace(dec_min - d_dec, dec_max + d_dec, subgrid_res * n_dec)

    ra_subgrid, dec_subgrid = np.meshgrid(ra_new, dec_new)
    return image2array(ra_subgrid), image2array(dec_subgrid)

def convert_bool_list(n, k=None):
    """
    returns a bool list of the length of the lens models
    if k = None: returns bool list with True's
    if k is int, returns bool list with False's but k'th is True
    if k is a list of int, e.g. [0, 3, 5], returns a bool list with True's in the integers listed and False elsewhere
    if k is a boolean list, checks for size to match the numbers of models and returns it

    :param n: integer, total lenght of output boolean list
    :param k: None, int, or list of ints
    :return: bool list
    """
    if k is None:
        bool_list = [True] * n
    elif isinstance(k, (int, np.integer)):  # single integer
        bool_list = [False] * n
        bool_list[k] = True
    elif len(k) == 0:  # empty list
        bool_list = [False] * n
    elif isinstance(k[0], bool):
        if n != len(k):
            raise ValueError('length of selected lens models in format of boolean list is %s '
                             'and does not match the models of this class instance %s.' % (len(k), n))
        bool_list = k
    elif isinstance(k[0], (int, np.integer)):  # list of integers
        bool_list = [False] * n
        for i, k_i in enumerate(k):
            if k_i is not False:
                # if k_i is True:
                #    bool_list[i] = True
                if k_i < n:
                    bool_list[k_i] = True
                else:
                    raise ValueError("k as set by %s is not convertable in a bool string!" % k)
    else:
        raise ValueError('input list k as %s not compatible' % k)
    return bool_list

def check_psd_force_symmetry(X_in):
    """ensure that the matrix X is symmetric positive-semidefinite"""
    X = np.copy(X_in)
    # check if positive-semidefinite
    if not np.all(np.linalg.eigvals(X) > 0):
        min_eig = np.min(np.real(np.linalg.eigvals(X)))
        print(f"correcting for negative eigenvalues with {10*min_eig}")
        X -= 10*min_eig * np.eye(*X.shape)

    # check symmetry (sometimes condition for being positive-semidefinite)
    if not np.all(X - X.T == 0):
        print("forcing symmetry")
        lower_triangle = np.tril(X)
        X = lower_triangle + lower_triangle.T - np.diag(lower_triangle.diagonal())

    # final check
    assert (np.all(np.linalg.eigvals(X) > 0) and np.all(X - X.T == 0))
    return X

def read_json(input_path):
    with open(input_path,'r') as f:
        input_str = f.read()
        input_str = re.sub(re.compile("/\*.*?\*/", re.DOTALL), "", input_str)
        input_str = re.sub(re.compile("//.*?\n" ), "", input_str)
        json_in   = json.loads(input_str)
    return json_in



# Defines a uniform profile
# 
# Copyright (c) 2021, herculens developers and contributors

__author__ = 'austinpeel', 'aymgal'


import jax.numpy as jnp


__all__ = ['Uniform']


class Uniform(object):
    """
    class for uniform light profile
    """
    param_names = ['amp']
    lower_limit_default = {'amp': -100}
    upper_limit_default = {'amp': 100}
    fixed_default = {key: False for key in param_names}

    def __init__(self):
        pass

    def function(self, x, y, amp):
        """

        :param x:
        :param y:
        :param sigma0:
        :param a:
        :param s:
        :param center_x:
        :param center_y:
        :return:
        """
        return jnp.ones_like(x) * amp



import setuptools
import os

name = 'herculens'

release_info = {}
infopath = os.path.abspath(os.path.join(os.path.dirname(__file__),
                           name, 'info.py'))
with open(infopath) as open_file:
    exec(open_file.read(), release_info)

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

# Use the README as long description but remove the second version 
# of the logo (readable only by GitHub's dark mode interface)
long_description = readme.replace('<img src="https://raw.githubusercontent.com/Herculens/herculens/main/images/horizontal_dark_bg.png#gh-dark-mode-only" width="600" alt="Herculens logo" />', '')

# Python version
python_requires = '>=3.10'

# Minimal required packages (see requirements.txt for versions)
install_requires = [
    'numpy',
    'scipy',
    'jax',
    'jaxlib',
    'objax',
    'scikit-image',
    # 'https://github.com/aymgal/utax.git',  # use requirements.txt
]

# Minimal optional packages (see requirements.txt for versions and complete list)
install_optional = [
    'matplotlib',    # for plotting
    'optax',         # for optimizers
    'jaxopt',        # for optimizers
    'findiff',       # for building matrix with finite difference coefficients
    'numpyro',       # for SVI and HMC sampling
    'blackjax',      # for JAX-based HMC sampling
]

version = release_info['__version__']

setuptools.setup(
    name=name,
    author=release_info['__author__'],
    author_email=release_info['__email__'],
    version=version,
    url=release_info['__url__'],
    download_url=f"https://github.com/Herculens/herculens/archive/refs/tags/v{version}.tar.gz",
    packages=setuptools.find_packages(),
    license=release_info['__license__'],
    description=release_info['__description__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=["herculens", "lensing", "gravitation", "astrophysics"],

    python_requires=python_requires,
    install_requires=install_requires,
    extras_require={
        "opt": install_optional,  # installable via `pip install herculens[opt]`
    },
    setup_requires=['pytest-runner',],
    tests_require=['pytest', 'pytest-cov', 'pytest-pep8'],
)



# Defines regularization choices
# 
# Copyright (c) 2023, herculens developers and contributors


__author__ = 'aymgal'


import numpy as np
from herculens.RegulModel.Methods import sparsity, constraints
from herculens.Util import util


__all__ = ['RegularizationModel']


SUPPORTED_MODELS = {
    'source': [
        'SPARSITY_STARLET',
        'SPARSITY_STARLET_2',
        'SPARSITY_BLWAVELET',
        'POSITIVITY',
    ],
    'lens_light': [
        'SPARSITY_STARLET',
        'SPARSITY_STARLET_2',
        'SPARSITY_BLWAVELET',
        'POSITIVITY',
    ],
    'lens_mass': [
        'SPARSITY_STARLET_POTENTIAL',
        'SPARSITY_STARLET_2_POTENTIAL',
        'SPARSITY_BLWAVELET_POTENTIAL',
        'POSITIVITY_POTENTIAL',
        'NEGATIVITY_POTENTIAL',
        'POSITIVITY_CONVERGENCE',
    ]
}


class RegularizationModel(object):

    """
    list items should be tuple (model_type, profile_index, method)
    e.g. method_list = [
        ('source', 0, 'SPARSITY_STARLET'),
        ('source', 0, 'POSITIVITY'),
        ('lens_mass', 0, 'SPARSITY_STARLET_2_POTENTIAL'),
    ]
    """

    _model_types = ['source', 'lens_light', 'lens_mass']

    def __init__(self, method_type_list):
        self.method_list, self.param_names = self._setup_methods(method_type_list)

    def initialize(self, lens_image, kwargs_params, **kwargs):
        for method in self.method_list:
            method.initialize(lens_image, kwargs_params, **kwargs)

    def log_prob(self, kwargs_params, kwargs_hyperparams):
        """
        Returns the log-probability term associated to regularization methods,
        given a set of (model) parameters and (regularization) hyperparameters
        """
        log_prob = 0.
        for i, method in enumerate(self.method_list):
            log_prob += method.log_prob(kwargs_params, **kwargs_hyperparams[i])
        return log_prob

    def get_weights(self):
        weights_list = []
        for method in self.method_list:
            if method.has_weights:
                weights_list.append(method.weights)
            else:
                weights_list.append(None)
        return weights_list

    @staticmethod
    def _setup_methods(method_type_list):
        method_list, param_names = [], []
        for model_type, profile_index, method_type in method_type_list:
            if model_type in ('source', 'lens_light'):
                if method_type == 'SPARSITY_STARLET':
                    method = sparsity.SparsityStarlet(model_type, profile_index)
                elif method_type == 'SPARSITY_STARLET_2':
                    method = sparsity.SparsityStarlet(model_type, profile_index,
                                                      second_gen=True)
                elif method_type == 'SPARSITY_BLWAVELET':
                    method = sparsity.SparsityBLWavelet(model_type, profile_index)
                elif method_type == 'POSITIVITY':
                    method = constraints.Positivity(model_type, profile_index)
            elif model_type == 'lens_mass':
                if method_type == 'SPARSITY_STARLET_POTENTIAL':
                    method = sparsity.SparsityStarlet(model_type, profile_index,
                                                     mass_form='potential')
                elif method_type == 'SPARSITY_STARLET_2_POTENTIAL':
                    method = sparsity.SparsityStarlet(model_type, profile_index,
                                                      mass_form='potential',
                                                      second_gen=True)
                elif method_type == 'SPARSITY_BLWAVELET_POTENTIAL':
                    method = sparsity.SparsityBLWavelet(model_type, profile_index,
                                                      mass_form='potential')
                elif method_type == 'POSITIVITY_POTENTIAL':
                    method = constraints.Positivity(model_type, profile_index,
                                                    mass_form='potential')
                elif method_type == 'NEGATIVITY_POTENTIAL':
                    method = constraints.Negativity(model_type, profile_index,
                                                    mass_form='potential')
                elif method_type == 'POSITIVITY_CONVERGENCE':
                    method = constraints.Positivity(model_type, profile_index,
                                                    mass_form='convergence')
            else:
                raise ValueError(f"Regularization method is unsupported "
                                 f"(supported model types are {self._model_types} "
                                 f"and supported methods are {SUPPORTED_MODELS[model_type]})")
            method_list.append(method)
            param_names.append(method.param_names)
        return method_list, param_names



# Handles data and model pixelated grids
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the Data module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


# TODO: combine with grid class from numerics

import copy
import numpy as np
from herculens.Coordinates.coord_transforms import Coordinates
from herculens.Util import util


__all__ = ['PixelGrid']



class PixelGrid(Coordinates):
    """
    class that manages a specified pixel grid (rectangular at the moment) and its coordinates
    """

    def __init__(self, nx, ny, transform_pix2angle=None, ra_at_xy_0=None, dec_at_xy_0=None):
        """

        :param nx: number of pixels in x-axis
        :param ny: number of pixels in y-axis
        :param transform_pix2angle: 2x2 matrix, mapping of pixel to coordinate
        :param ra_at_xy_0: ra coordinate at pixel (0,0)
        :param dec_at_xy_0: dec coordinate at pixel (0,0)
        """
        base_pix_scl = 1.  # some default pixel size
        if transform_pix2angle is None:
            transform_pix2angle = base_pix_scl * np.eye(2)  # transformation matrix pixel <-> angle
        if ra_at_xy_0 is None or dec_at_xy_0 is None:
            half_size_nx = nx * base_pix_scl / 2
            ra_at_xy_0 = - half_size_nx + base_pix_scl / 2.  # position of the (0, 0) with respect to bottom left pixel
            half_size_ny = ny * base_pix_scl / 2
            dec_at_xy_0 = - half_size_ny + base_pix_scl / 2.  # position of the (0, 0) with respect to bottom left pixel
            
        super(PixelGrid, self).__init__(transform_pix2angle, ra_at_xy_0, dec_at_xy_0)
        self._nx = nx
        self._ny = ny
        self._x_grid, self._y_grid = self.coordinate_grid(nx, ny)
        self._model_grids = {}

    @property
    def num_pixel(self):
        """

        :return: number of pixels in the data
        """
        return self._nx * self._ny

    @property
    def num_pixel_axes(self):
        """

        :return: number of pixels per axis, nx ny
        """
        return self._nx, self._ny

    @property
    def width(self):
        """

        :return: width of data frame
        """
        return self._nx * self.pixel_width, self._ny * self.pixel_width

    @property
    def center(self):
        """

        :return: center_x, center_y of coordinate system
        """
        return np.mean(self._x_grid), np.mean(self._y_grid)

    def shift_coordinate_system(self, x_shift, y_shift, pixel_unit=False):
        """
        shifts the coordinate system
        :param x_shift: shift in x (or RA)
        :param y_shift: shift in y (or DEC)
        :param pixel_unit: bool, if True, units of pixels in input, otherwise RA/DEC
        :return: updated data class with change in coordinate system
        """
        self._shift_coordinates(x_shift, y_shift, pixel_unit=pixel_unit)
        self._x_grid, self._y_grid = self.coordinate_grid(self._nx, self._ny)

    @property
    def pixel_coordinates(self):
        """

        :return: RA coords, DEC coords
        """
        return self._x_grid, self._y_grid

    @property
    def pixel_axes(self):
        """

        :return: RA coords, DEC coords
        """
        return self._x_grid[0, :], self._y_grid[:, 0]

    @property
    def extent(self):
        x_coords, y_coords = self.pixel_axes
        return [x_coords[0], x_coords[-1], y_coords[0], y_coords[-1]]
    
    @property
    def plt_extent(self):
        """set of coordinates of the borders of the grid (useful for matplotlib functions)"""
        extent = copy.copy(self.extent)
        # WARNING: the following assumes NO ROTATION (i.e. coordinates axes aligned with x/y axes)
        pix_scl_x = self._Mpix2a[0, 0]
        pix_scl_y = self._Mpix2a[1, 1]
        if self.x_is_inverted:
            extent[0] += pix_scl_x / 2.
            extent[1] -= pix_scl_x / 2.
        else:
            extent[0] -= pix_scl_x / 2.
            extent[1] += pix_scl_x / 2.
        if self.y_is_inverted:
            extent[2] += pix_scl_y / 2.
            extent[3] -= pix_scl_y / 2.
        else:
            extent[2] -= pix_scl_y / 2.
            extent[3] += pix_scl_y / 2.
        return extent

    def create_model_grid(self, num_pixels=None, pixel_scale_factor=None, grid_center=None, grid_shape=None):
        """
        :param num_pixels: number of pixels on each side. 
        Currently this only works with square grids, so grid_shape needs to setup a square grid.
        :param pixel_scale_factor: multiplicative factor to go from original pixel width to new pixel width.
        If None, defaults to the 1. If num_pixels is provided, pixel_scale_factor is be ignored.
        :param grid_center: 2-tuple (center_x, center_y) with grid center in angular units
        If None, defaults to the original grid center. 
        :param grid_shape: 2-tuple (width, height) window size in angular units
        If None, defaults to the original window size.
        """
        unchanged_count = 0
        if grid_center is None or grid_center == self.center:
            grid_center_ = self.center
            unchanged_count += 1
        else:
            grid_center_ = grid_center
        if grid_shape is None or grid_shape == self.width:
            grid_shape_ = self.width
            unchanged_count += 1
        else:
            grid_shape_ = grid_shape
        if ((num_pixels is None and (pixel_scale_factor is None or pixel_scale_factor == 1)) or
            (num_pixels is not None and (num_pixels, num_pixels) == self.num_pixel_axes)):
            pixel_scale_factor_ = 1
            num_pixels_ = None
            unchanged_count += 1
        else:
            pixel_scale_factor_ = pixel_scale_factor
            num_pixels_ = num_pixels

        # in case it's the same region as the base coordinate grid
        if unchanged_count == 3:
            return copy.deepcopy(self)

        center_x, center_y = grid_center_
        width, height = grid_shape_
        if num_pixels_ is None:
            pixel_width = self.pixel_width * float(pixel_scale_factor_)
            nx = round(width / pixel_width)
            ny = round(height / pixel_width)
        else:
            nx = ny = num_pixels_  # assuming square grid
            if height != width:
                raise ValueError(f"Setting number of side pixels only works with square grids "
                                 f"(grid shape {grid_shape_} was provided).")
            pixel_width = width / float(nx)

        transform_pix2angle = self.transform_pix2angle / self.pixel_width * pixel_width

        cx, cy = nx / 2., ny / 2.
        cra, cdec = transform_pix2angle.dot(np.array([cx, cy]))
        ra_at_xy_0, dec_at_xy_0 = - cra + center_x + pixel_width/2., - cdec + center_y + pixel_width/2.

        return PixelGrid(nx, ny, transform_pix2angle, ra_at_xy_0, dec_at_xy_0)



import jax.numpy as jnp
import jax.scipy.stats as jstats

try:
    import nifty8.re as jft
except ImportError:
    raise ImportError("The nifty8.re module is not installed. "
                      "Please install NIFTy8 (see https://github.com/NIFTy-PPL/NIFTy) "
                      "in order to use the CorrelatedField class "
                      "or the jifty_util submodule of Herculens.")
from nifty8.re.tree_math import ShapeWithDtype


def prepare_correlated_field(
        key, num_pix_xy, border_xy, 
        kwargs_amplitude, 
        kwargs_fluctuations,
        kernel_type='powerlaw',
        correlation_type='power',
        matern_renorm_amp=False,
        num_pix_wl=1, 
        border_wl=0, 
        kwargs_fluctuations_wl=None,
        field_key='field',
        param_key_xy='xy_dim', 
        param_key_wl='wl_dim',
        correlation_type_wl='none',
        non_linearity='exp'
    ):
    """Utility that sets up a nifty.re correlated field forward model.

    Parameters
    ----------
    key : str
        Suffix (like a parameter name) added to the internal field parameter names
    num_pix_xy : int
        Number of pixels along each spatial dimensions (i.e. we assume a square grid).
    border_xy : int
        Number of pixels to crop on each sides of the field before it's returned by the model.
    kwargs_amplitude : dict
        Parameters of the field (see nifty.re documentation)
    kwargs_fluctuations : _type_
        Parameters of the field (see nifty.re documentation)
    kernel_type : str, optional
        Type of kernel to use for the field, either 'powerlaw' or 'matern', by default 'powerlaw'.
        If 'matern', the kwargs_fluctuations are given to the CorrelatedFieldMaker's
        method `add_fluctuations_matern()`, instead of `add_fluctuations()`.
        See the nifty.re documentation for more details. Note that the 'matern' kernel
        is only supported for the spatial dimensions, not for the spectral dimensions.
    num_pix_wl : int, optional
        Same as num_pix_xy but for the spectral (i.e. along wavelengths) dimensions, by default 1.
    border_wl : int, optional
        Same as border_xy but for the spectral (i.e. along wavelengths) dimension, by default 1.
    kwargs_fluctuations_wl : _type_, optional
        Same as kwargs_fluctuations for the spectral (i.e. along wavelengths) dimension
    non_linearity : str, optional
        Whether or not to apply a non-linearity to get specific behaviors, by default 'exp'.
        For instance, exponential non-linearity is useful to ensure non-negative values.

    Returns
    -------
    tuple of (jft.CorrelatedFieldMaker, jft.Model, int, int)
        Instance of the correlated field maker, the forward field model, 
        the total number of pixels in the spatial dimensions, 
        and the total number of pixels in the spectral dimensions.

    Raises
    ------
    ValueError
        If the non-linearity is not implemented.
    """
    if non_linearity not in ['exp', 'none']:
        raise ValueError(f"Non-linearity '{non_linearity}' not implemented.")
    if kernel_type not in ['powerlaw', 'matern']:
        raise ValueError(f"Kernel type '{kernel_type}' not supported. Supported types are 'powerlaw' and 'matern'.")
    
    # Correlated field parameters
    cfm = jft.CorrelatedFieldMaker(key + f'_{field_key}_')
    cfm.set_amplitude_total_offset(**kwargs_amplitude)

    # Setup the power-spectrum in the spectral dimension (if multi-band fitting)
    if num_pix_wl > 1:
        num_pix_wl_tot = num_pix_wl + 2 * border_wl
        cfm.add_fluctuations(
            [num_pix_wl_tot],
            distances=1./num_pix_wl_tot,  # only makes a difference to get proper units for wavenumbers and certain types of covariance kernels
            **kwargs_fluctuations_wl,
            prefix=param_key_wl+'_',  # prefix key to e.g. distinguish between multiple fields along different dimensions
            non_parametric_kind=correlation_type_wl,
            harmonic_type='fourier',
        )
    else:
        num_pix_wl_tot = 1

    # Setup the power-spectrum in the spatial dimensions
    num_pix_xy_tot = num_pix_xy + 2 * border_xy
    if kernel_type == 'powerlaw':
        cfm.add_fluctuations(
            [num_pix_xy_tot, num_pix_xy_tot],
            distances=1./num_pix_xy_tot,  # only makes a difference to get proper units for wavenumbers and certain types of covariance kernels
            **kwargs_fluctuations,
            prefix=param_key_xy+'_',  # prefix key to e.g. distinguish between multiple fields along different dimensions
            non_parametric_kind=correlation_type,
            harmonic_type='fourier',
        )
    elif kernel_type == 'matern':
        cfm.add_fluctuations_matern(
            [num_pix_xy_tot, num_pix_xy_tot],
            distances=1./num_pix_xy_tot,  # only makes a difference to get proper units for wavenumbers and certain types of covariance kernels
            renormalize_amplitude=matern_renorm_amp,
            **kwargs_fluctuations,
            prefix=param_key_xy+'_',  # prefix key to e.g. distinguish between multiple fields along different dimensions
            non_parametric_kind=correlation_type,
            harmonic_type='fourier',
        )
    # NOTE: in the pure nifty version, the distances are also 1/num_pix in the SimpleCorrelatedField model

    # Finalize the correlated field
    field_transform = cfm.finalize()

    # Apply non-linear transformation to get the final light model
    if non_linearity == 'exp':
        field_transform = ExpCroppedFieldTransform(key, field_transform, border_xy, border_wl)
        
    return cfm, field_transform, num_pix_xy_tot, num_pix_wl_tot


def prepare_light_correlated_field(*args, **kwargs):
    # just for backwards compatibility
    return prepare_correlated_field(*args, **kwargs)


def concatenate_fields(*fields):
    def operator(x):
        op = {}
        for field in fields:
            op.update(field(x))
        return op
    domain = {}
    for field in fields:
        domain.update(field.domain)
    return jft.Model(
        lambda x: operator(x),
        domain=domain,
    )


def NormalTransform(m, s, key):
    """m: mean, s: std"""
    print(f"Normal prior for '{key}' ({m}, {s})")
    if not hasattr(m, 'shape'):
        shape = ()
    else:
        shape = m.shape
    return jft.Model(
        lambda x: {key: x[key]*s + m},
        domain={key: ShapeWithDtype(shape, jnp.float64)},
    )

def NormalTransform_inverse_func(x, m, s):
    """m: mean, s: std, to go some normal distribution to standard normal distribution"""
    return (x - m) / s

def LognormalTransform(m, s, key):
    """m: mean, s: std"""
    print(f"Log-normal prior for '{key}' ({m}, {s})")
    log_s = jnp.sqrt(jnp.log1p((s/m)**2))
    log_m = jnp.log(m) - log_s**2/2.
    if not hasattr(m, 'shape'):
        shape = ()
    else:
        shape = m.shape
    return jft.Model(
        lambda x: {key: jnp.exp(x[key]*log_s + log_m)},
        domain={key: ShapeWithDtype(shape, jnp.float64)},
    )

def UniformTransform(l, h, key):
    """l: low, h: high"""
    print(f"Uniform prior for '{key}' [{l}, {h}]")
    if not hasattr(l, 'shape'):
        shape = ()
    else:
        shape = l.shape
    return jft.Model(
        lambda x: {key: (h - l) * jstats.norm.cdf(x[key]) + l},
        domain={key: ShapeWithDtype(shape, jnp.float64)},
    )

def UniformTransform_inverse_func(x, l, h):
    """m: mean, s: std, to go from uniform distribution to standard normal distribution"""
    return jstats.norm.ppf( (x - l) / (h - l) )

def ExpCroppedFieldTransform(key, cf, bxy, bwl):
    print(f"Exponential Cropped correlated field prior for '{key}'")
    def crop(x):
        x_ = cf(x)
        if bwl == 0:
            if bxy > 0:
                x_ = x_[bxy:-bxy, bxy:-bxy]
        else:
            x_ = x_[bwl:-bwl, :, :]
            if bxy > 0:
                x_ = x_[:, bxy:-bxy, bxy:-bxy]
        return x_
    return jft.Model(
        lambda x: {key: jnp.exp(crop(x))},  # here the target key is different from domain keys below
        domain=cf.domain,
    )



"""

This file initializes the coolest module
and provides some basic information about the package.

"""

# Set the package release version
version_info = (0, 2, 3)
__version__ = '.'.join(str(c) for c in version_info)

# Set the package details
__author__ = 'Aymeric Galan, Austin Peel, Giorgos Vernardos & Herculens contributors'
__email__ = 'aymeric.galan@gmail.com'
__year__ = '2021-2025'
__credits__ = 'Herculens contributors, EPFL (see AUTHORS.md)'
__url__ = 'https://github.com/Herculens/herculens'
__description__ = 'Auto-differentiable strong lens modelling'
__license__ = 'MIT'



# JAX implementation of the PIEMD profile for the lens modeling code GLEE.
#
# This is based on the GLEE implementation see Suyu & Halkola (2010) 
# and Suyu et al. (2012), which was ported to JAX in Wang et al. (2025). 
#
# This module is largely based on https://github.com/HanWang2021/Lens-Mass-Profile-GLaD-.
# The main change from the original code above is that lensing methods (e.g. get_deflection_angle)
# take as arguments the coordinates, rather than fixing those in the constructor.
# All methods specific to the use of the profile in the GLaD code have been removed as well.
#
# Copyright (c) 2025: Han Wang (MPA), Aymeric Galan (MPA)
# Copyright (c) 2023: herculens developers and contributors
# Copyright (c) 2010: Sherry Suyu, Aleksi Halkola (MPA)


__author__ = 'HanWang2021', 'aymgal'


from functools import partial
import numpy as np
import jax
import jax.numpy as jnp
from jax.tree_util import register_pytree_node



def rotation(xx, yy, phi):
    cos_phi, sin_phi = jnp.cos(phi), jnp.sin(phi)
    xx_rot = cos_phi*xx + sin_phi*yy
    yy_rot = -sin_phi*xx + cos_phi*yy
    return xx_rot, yy_rot


class PIEMD(object):
    """
    Pseudo isothermal elliptical mass distribution (PIEMD), based on the
    implementation in the GLEE modelling code.
    
    The original GLEE parametrization is

    kappa(x,y) = E0 / (2*Sqrt(w^2 + rem^2)),
    where rem^2 = x^2/(1+e)^2 + y^2/(1-e)^2,

    with parameters:
    E0 is the lens strength,
    w is the core radius, and
    e is the ellipticity, e=(1.-q)/(1.+q)   [inverting gives q = (1-e)/(1+e)]

    In the limiting case where w=e=0, E0 is the Einstein radius of the SIS.

    If `scale_flag=False`, then E0 = thetaE

    If `scale_flag=True`, then E0 = thetaE, E0 = E0*E0 / (sqrt(E0 * E0 + w * w) - w)

    If (w < r_soft) then w will be set to `r_soft`.

    In the Herculens-adapted code below, we use theta_E parameter for E0 (or Elim), phi for the position angle, and q for the axis ratio, 
    and r_core for the core radius w.
    The 'scale' flag and softening radius are set at construction time.

    Parameters
    ----------
    r_soft : _type_, optional
        Softening radius, by default 1e-8
    scale_flag : bool, optional
        'scale' flag usually given in GLEE config files, by default True, 
        which leads to theta_E being simply the lensing strength.
    """

    param_names = ['theta_E', 'r_core', 'q', 'phi', 'center_x', 'center_y']
    lower_limit_default = {'theta_E': 0, 'r_core': 0, 'q': 0.2, 'phi': -np.pi/2., 'center_x': -100, 'center_y': -100}
    upper_limit_default = {'theta_E': 10, 'r_core': 1e10, 'q': 1., 'phi': +np.pi/2., 'center_x': 100, 'center_y': 100}
    fixed_default = {key: False for key in param_names}

    def __init__(self, r_soft=1e-8, scale_flag=True):
        self._r_soft = r_soft
        self._scale_flag = scale_flag
    
    @partial(jax.jit, static_argnums=(0,))
    def function(self, x, y, theta_E, r_core, q, phi, center_x=0, center_y=0):  # flags should be also an array
        # return lens potential
        e = (1.-q)/(1.+q)
        w = r_core
        theta_E = jnp.where(self._scale_flag, theta_E**2/(jnp.sqrt(theta_E*theta_E + w * w) - w), theta_E)
        return jax.lax.cond(
            e != 0, 
            self._potential_elliptical, 
            self._potential_spherical, 
            x, y, center_x, center_y, e, phi, theta_E, w
        )
    
    @partial(jax.jit, static_argnums=(0,))
    def derivatives(self, x, y, theta_E, r_core, q, phi, center_x=0, center_y=0):
        # return deflection angle
        e = (1.-q)/(1.+q)
        w = r_core
        theta_E = jnp.where(self._scale_flag, theta_E**2/(jnp.sqrt(theta_E*theta_E + w * w) - w), theta_E)
        return jax.lax.cond(
            e!=0, 
            self._deflection_elliptical, 
            self._deflection_spherical, 
            x, y, center_x, center_y, e, phi, theta_E, w
        )
    
    @partial(jax.jit, static_argnums=(0,))
    def hessian(self, x, y, theta_E, r_core, q, phi, center_x=0, center_y=0):
        e = (1.-q)/(1.+q)
        w = r_core
        theta_E = jnp.where(self._scale_flag, theta_E**2/(jnp.sqrt(theta_E*theta_E + w * w) - w), theta_E)
        return jax.lax.cond(
            e!=0, 
            self._hessian_elliptical, 
            self._hessian_spherical,
            x, y, center_x, center_y, e, phi, theta_E, w
        )
    
    @partial(jax.jit, static_argnums=(0,))
    def kappa(self, x, y, theta_E, r_core, q, phi, center_x=0, center_y=0):
        xx = x - center_x
        yy = y - center_y
        xx, yy = rotation(xx, yy, phi)
        r2 = xx**2 + yy**2/(q*q)
        w = r_core
        theta_E = jnp.where(self._scale_flag, theta_E**2/(jnp.sqrt(theta_E*theta_E + w * w) - w), theta_E)
        kap = (theta_E/(1+q))*jnp.sqrt((1/(4*w**2/(1+q)**2 + r2)))
        return kap
    
    @partial(jax.jit, static_argnums=(0,))
    def _potential_spherical(self, x, y, x_centre, y_centre, _, pa, theta_E, w):
        # potential, given q = 1
        xx = x - x_centre
        yy = y - y_centre
        xx, yy = rotation(xx, yy, pa)
        rm = jnp.sqrt((xx)**2 + (yy)**2)
        phi = theta_E *(jnp.sqrt(rm*rm+w*w) - w*jnp.log(w+jnp.sqrt(rm*rm+w*w)) - w + w*jnp.log(2.*w) )
        return phi
    
    @partial(jax.jit, static_argnums=(0,))
    def _potential_elliptical(self, x, y, x_centre, y_centre, e, pa, theta_E, w):
        # potential, given q != 1
        xx = x - x_centre
        yy = y - y_centre
        xx, yy = rotation(xx, yy, pa)
        #jax.debug.print("theta_E: {}", theta_E)
        rem = jnp.sqrt(xx*xx/((1+e)*(1+e))+yy*yy/((1-e)*(1-e)))
        rm = jnp.sqrt(xx*xx + yy*yy)
        sang = yy/rm
        cang = xx/rm
        eta = -0.5*jnp.arcsinh((2.*jnp.sqrt(e)/(1.-e)) * sang) + 0.5j * jnp.arcsin((2.*jnp.sqrt(e)/(1.+e)) * cang)
        zeta = 0.5*jnp.log((rem + jnp.sqrt(rem*rem + w*w))/w) + 0.0j
        cosheta = jnp.cosh(eta)
        coshplus = jnp.cosh(eta+zeta)
        coshminus = jnp.cosh(eta-zeta)
        Kstar = jnp.sinh(2.*eta) * jnp.log(cosheta*cosheta/(coshplus*coshminus)) + jnp.sinh(2.*zeta) * jnp.log(coshplus/coshminus)
        phi = (theta_E*w*(1.-e*e)/(2.*rem*jnp.sqrt(e))) * jnp.imag((xx - 1j*yy) * Kstar)
        return phi

    def _deflection_spherical(self, x, y, x_centre, y_centre, e, pa, theta_E, w):
        # return deflection_angle, q = 1, e is not used here, passing it as input is only because of the syntax of jax.lax.cond()
        tol = 1.e-10
        xx =  x - x_centre
        yy =  y - y_centre
        xx, yy = rotation(xx, yy, pa)
        commen_factor = jnp.sqrt(w*w + xx*xx + yy*yy)/(xx*xx+yy*yy)
        condition  = (xx*xx+yy*yy)/(w*w)
        alp_x = jnp.where(condition>=tol, theta_E*xx*((-w/(xx*xx+ yy*yy) + commen_factor)), 0.5*theta_E*(xx)/w)
        alp_y = jnp.where(condition>=tol, theta_E*yy*(-(w/(xx*xx + yy*yy))+ commen_factor),  0.5*theta_E*(yy)/w)
        alp_x, alp_y = rotation(alp_x, alp_y, -pa)
        return alp_x, alp_y

    def _deflection_elliptical(self, x, y, x_centre, y_centre, e, pa, E0, w):
        # return deflection_angle, given q!=1
        dx =  x - x_centre
        dy =  y - y_centre
        dx, dy = rotation(dx, dy, pa)
        Istar = ((-0.5j) * (1 - (e * e)) * E0 * jnp.log((((1 - e) * dx) / (1 + e) - (1j * (1 + e) * dy) / (1 - e) + (2j) * jnp.sqrt(e) * jnp.sqrt(w * w + (dx * dx) / ((1 + e) * (1 + e)) + (dy * dy) / ((1 - e) * (1 - e)))) / ((2j) * jnp.sqrt(e) * w + dx - (1j) * dy))) / jnp.sqrt(e)
        alp_x, alp_y = jnp.real(Istar), jnp.imag(Istar)
        alp_x, alp_y = rotation(alp_x, alp_y, -pa)
        return alp_x, alp_y
    
    def _hessian_spherical(self, x, y, x_centre, y_centre, e, pa ,theta_E, w):
        xx =  x - x_centre
        yy =  y - y_centre
        dx, dy = rotation(xx, yy, pa)
        psi11 = (theta_E*(w*w*(-(dx)*(dx) + (dy)*(dy)) +(dy)*(dy)*((dx)*(dx) + (dy)*(dy)) +
		          w*((dx) - (dy))*((dx) + (dy))*jnp.sqrt(w*w + (dx)*(dx) + (dy)*(dy))))/(jnp.power((dx)*(dx) + (dy)*(dy),2)*jnp.sqrt(w*w + (dx)*(dx) + (dy)*(dy)))

        psi22 = (theta_E*(w*w*((dx) - (dy))*((dx) + (dy)) + (dx)*(dx)*((dx)*(dx) + (dy)*(dy)) + w*(-(dx)*(dx) + (dy)*(dy))* jnp.sqrt(w*w + (dx)*(dx) + (dy)*(dy))))/(jnp.power((dx)*(dx) + (dy)*(dy),2)*jnp.sqrt(w*w + (dx)*(dx) + (dy)*(dy)))
        psi12 = -((theta_E*(dx)*(dy)*((dx)*(dx) + (dy)*(dy) + 2*w*(w - jnp.sqrt(w*w + (dx)*(dx) + (dy)*(dy)))))/(jnp.power((dx)*(dx) + (dy)*(dy),2)*jnp.sqrt(w*w + (dx)*(dx) + (dy)*(dy))))
        return psi11, psi22, psi12

    def _hessian_elliptical(self, x, y, x_centre, y_centre, e, pa ,theta_E, w):
        xx =  x - x_centre
        yy =  y - y_centre
        dx, dy = rotation(xx, yy, pa)

        psi11= jnp.imag(((1 - (e*e))*theta_E*((2.j)*jnp.sqrt(e)*w + (dx) - (1.j)*(dy))*
		            (((1 - e)/(1 + e) + ((2.j)*jnp.sqrt(e)*(dx))/
		             (((1+e)*(1+e))*jnp.sqrt(w*w + (dx)*(dx)/((1+e)*(1+e)) +
		                    (dy)*(dy)/((1-e)*(1-e)))))/
		            ((2.j)*jnp.sqrt(e)*w + (dx) - (1.j)*(dy)) -
		             (((1 - e)*(dx))/(1 + e) - ((1.j)*(1 + e)*(dy))/(1 - e) +
		             (2.j)*jnp.sqrt(e)*
	               jnp.sqrt(w*w + (dx)*(dx)/((1+e)*(1+e)) +
		                  (dy)*(dy)/((1-e)*(1-e))))/
	              jnp.power((2.j)*jnp.sqrt(e)*w + (dx) - (1.j)*(dy),2)))/
		          (jnp.sqrt(e)*(((1 - e)*(dx))/(1 + e) -
		                    ((1.j)*(1 + e)*(dy))/(1 - e) +
		                    (2.j)*jnp.sqrt(e)*
		                  jnp.sqrt(w*w + (dx)*(dx)/((1+e)*(1+e)) +
	                          (dy)*(dy)/((1-e)*(1-e))))))/2.

        psi22=-jnp.real(((1 - (e*e))*theta_E*(2.j*jnp.sqrt(e)*w + (dx) - 1.j*(dy))*
		              ((((-1.j)*(1 + e))/(1 - e) +
		                (2.j*jnp.sqrt(e)*(dy))/
		               (((1-e)*(1-e))*
		                jnp.sqrt(w*w + (dx)*(dx)/((1+e)*(1+e)) +
		                      (dy)*(dy)/((1-e)*(1-e)))))/
		            (2.j*jnp.sqrt(e)*w + (dx) - 1.j*(dy)) +
	                 (1.j*(((1 - e)*(dx))/(1 + e) -
	                  (1.j*(1 + e)*(dy))/(1 - e) + 2.j*jnp.sqrt(e)*
		                   jnp.sqrt(w*w + (dx)*(dx)/((1+e)*(1+e)) +
	                        (dy)*(dy)/((1-e)*(1-e)))))/
	              jnp.power(2.j*jnp.sqrt(e)*w + (dx) - 1.j*(dy),2)))/
		          (jnp.sqrt(e)*(((1 - e)*(dx))/(1 + e) -
		                      (1.j*(1 + e)*(dy))/(1 - e) +
	                      2.j*jnp.sqrt(e)*jnp.sqrt(w*w + (dx)*(dx)/((1+e)*(1+e)) +
		                           (dy)*(dy)/((1-e)*(1-e))))))/2.

        psi12=-jnp.real(((1 - (e*e))*theta_E*(2.j*jnp.sqrt(e)*w + (dx) - 1.j*(dy))*
		             (((1 - e)/(1 + e) + (2.j*jnp.sqrt(e)*(dx))/
		                (((1+e)*(1+e))*jnp.sqrt(w*w + (dx)*(dx)/((1+e)*(1+e)) +
		                     (dy)*(dy)/((1-e)*(1-e)))))/
		              (2.j*jnp.sqrt(e)*w + (dx) - 1.j*(dy)) -
		              (((1 - e)*(dx))/(1 + e) - (1.j*(1 + e)*(dy))/(1 - e) +
	               2.j*jnp.sqrt(e)*jnp.sqrt(w*w + (dx)*(dx)/((1+e)*(1+e)) +
		                    (dy)*(dy)/((1-e)*(1-e))))/
		             jnp.power(2.j*jnp.sqrt(e)*w + (dx) - 1.j*(dy),2)))/
		 	   (jnp.sqrt(e)*(((1 - e)*(dx))/(1 + e) -
	                       (1.j*(1 + e)*(dy))/(1 - e) +
	                       2.j*jnp.sqrt(e)*jnp.sqrt(w*w + (dx)*(dx)/((1+e)*(1+e)) +
	                        (dy)*(dy)/((1-e)*(1-e))))))/2.
        return psi11, psi22, psi12

def flatten_func(self):
    children = ()
    aux_data = {
        'r_soft': self._r_soft,
        'scale_flag': self._scale_flag,
    }
    return (children, aux_data)

def unflatten_func(aux_data, children):
    # Here we avoid `__init__` because it has extra logic we don't require:
    obj = object.__new__(PIEMD)
    obj._r_soft = aux_data['r_soft']
    obj._flag_dpie = aux_data['scale_flag']
    return obj

register_pytree_node(PIEMD, flatten_func, unflatten_func)



# Defines the data noise model
# 
# Copyright (c) 2021, herculens developers and contributors
# Copyright (c) 2018, Simon Birrer & lenstronomy contributors
# based on the Data module from lenstronomy (version 1.9.3)

__author__ = 'sibirrer', 'austinpeel', 'aymgal'


import numpy as np
from functools import partial
import jax.numpy as jnp
from jax import random, jit
from herculens.Util import image_util


__all__ = ['Noise']


class Noise(object):
    """Class that builds a noise model, to be used in LensImage.

    Three possibilities exist, depending on what the user provides to the constructor:
    - a fixed noise map, which does not depend on any model-predicted flux.
    - an exposure time (or map) and a fixed background noise estimate,
    which will be used to estimate the noise based on an image model (see LensImage.model()).
    - an exposure time (or map) but no background noise estimate. In this case,
    the user should provide a (possibly varying) background_rms value when 
    calling the Noise.C_D_model().

    Parameters
    ----------
    nx : int
        number of data pixels along the x direction.
    ny : int
        number of data pixels along the y direction.
    exposure_time : np.array or float, optional
        exposure time, either common for all pixels or for each individual 
        data pixel. By default None
    background_rms : float, optional
        Root-mean-square value (standard deviation) of Gaussian background noise. 
        By default None.
    noise_map : np.array or int, optional
        noise standard of each individual pixel.
        If provide, overwrites any value of background_rms and exposure_time. 
        By default None.
    variance_boost_map : np.array, optional
        fixed (not model-dependent) variance boost map. By default None.
    verbose : bool, optional
        If True, outputs warning message at construction. By default True.

    Raises
    ------
    ValueError
        If neither a noise map or exposure time is provided.
    """

    def __init__(self, nx, ny, exposure_time=None, background_rms=None, 
                 noise_map=None, variance_boost_map=None, verbose=True):
        self._data = None  # TODO: is that really useful?
        self._nx, self._ny = nx, ny  # TODO: is that really useful?
        if noise_map is not None:
            assert np.shape(noise_map) == (nx, ny)
            assert np.all(noise_map > 0.)
        if exposure_time is None and noise_map is None:
            raise ValueError("Either a fixed noise map or an exposure time "
                             "(or exposure map) should be provided.")
        self._noise_map = noise_map
        if exposure_time is not None:
            # make sure no negative exposure values are present no dividing by zero
            if isinstance(exposure_time, int) or isinstance(exposure_time, float):
                if exposure_time <= 1e-10:
                    exposure_time = 1e-10
            else:
                exposure_time[exposure_time <= 1e-10] = 1e-10
        self._exp_map = exposure_time
        self._background_rms = background_rms
        if (self._noise_map is None and self._background_rms is None
            and verbose is True):
            print("Warning: both `noise_map` and `background_rms` are None; "
                  "`background_rms` should then be given as a model parameter "
                  "to estimate the noise variance via C_D_model().")
        if variance_boost_map is None:
            variance_boost_map = np.ones((self._nx, self._ny))
        self.global_boost_map = variance_boost_map

    def set_data(self, data):
        assert np.shape(data) == (self._nx, self._ny)
        self._reset_cache()
        self._data = data

    def compute_noise_map_from_model(self, model, as_jax_array=True):
        if self._noise_map is not None:
            UserWarning("Previous noise map will be replaced with new estimate from a model")
            self._noise_map = None
            #raise ValueError("A noise map has already been set!")
        noise_map = jnp.sqrt(self.C_D_model(model))
        self._reset_cache()
        self._noise_map = noise_map if as_jax_array else np.array(noise_map)

    def realisation(self, model, prng_key, add_background=True, add_poisson_model=True):
            """Draws a noise realization consistent with the model,
            based on the specified parameters.

            Parameters
            ----------
            model : object
                The input model to which noise realizations will be added.
            prng_key : jax.random.PRNGKey
                The random key used for generating random numbers.
            add_background : bool, optional
                Whether to add background noise to the model. Default is True.
            add_poisson_model : bool, optional
                Whether to add Poisson noise (shot noise) to the model. Default is True.

            Returns
            -------
            float
                The total noise realization added to the model.

            Raises
            ------
            ValueError
                If `add_poisson_model` is True but `exposure_map` is None.
            ValueError
                If `add_background` is True but `background_rms` is None.
            """
            noise_real = 0.
            key1, key2 = random.split(prng_key)
            if add_poisson_model:
                if self.exposure_map is None:
                    raise ValueError("An exposure time (or map) is needed to add Poisson (shot) noise")
                noise_real += image_util.add_poisson(model, self.exposure_map, key1)
            if add_background:
                if self.background_rms is None:
                    raise ValueError("A background RMS value is needed to add background noise")
                noise_real += image_util.add_background(model, self.background_rms, key2)
            return noise_real
    
    @property
    def variance_boost_map(self):
        # NOTE: we use a setter for backward compatibility reasons
        return self.global_boost_map
    
    @variance_boost_map.setter
    def variance_boost_map(self, boost_map):
        self.global_boost_map = boost_map

    @property
    def background_rms(self):
        """
        The standard deviation ("RMS") of the background noise.
        
        NOTE: "RMS" is a misleading term; it will be changed to sigma 
        or standard deviation in the future.

        Returns
        -------
        float
            Standard deviation of the background noise

        Raises
        ------
        ValueError
            If neither a background_rms value nor a noise map is available. 
        """
        if self._background_rms is None:
            if self._noise_map is None:
                raise ValueError("rms background value as 'background_rms' not specified!")
            else:
                print("Warning: Estimating the background RMS by the median of the noise map.")
            self._background_rms = np.median(self._noise_map)
        return self._background_rms

    @property
    def exposure_map(self):
        """
        Units of data and exposure map should result in:
        number of flux counts = data * exposure_map

        Returns
        -------
        _type_
            exposure map for each pixel

        Raises
        ------
        ValueError
            If neither an exposure time (or map) nor a noise map is available.
        """
        if self._exp_map is None:
            if self._noise_map is None:
                raise ValueError("Exposure map has not been specified in Noise() class!")
        return self._exp_map

    @property
    def C_D(self):
        """Covariance matrix of all pixel values in 2d numpy array (only diagonal component)
        The covariance matrix is estimated from the data (not from any model).
        WARNING: For low count statistics, the noise in the data may lead to biased estimates of the covariance matrix.

        Returns
        -------
        jax.numpy.array
            Noise variance per pixel.

        Raises
        ------
        ValueError
            If no data image nor a noise map is available.
        """
        if not hasattr(self, '_C_D'):
            if self._noise_map is not None:
                self._C_D = self._noise_map ** 2
            else:
                if self._data is None:
                    raise ValueError("No imaging data array has been set, impossible to estimate the diagonal covariance matrix")
                self._C_D = self.total_variance(self._data, self.background_rms, self.exposure_map)
        return self._C_D

    @partial(jit, static_argnums=(0, 4))
    def C_D_model(self, model, background_rms=None, boost_map=1., 
                  force_recompute=False):
        """Returns the estimate of the variance per pixel (i.e. the diagonal
        of the data covariance matrix) with contributions from background noise
        and shot noise from the model-predicted flux.

        Parameters
        ----------
        model : jax.numpy.array
            Image model for shot noise (Poisson noise) estimation.
        background_rms : float, optional
            Standard deviation of the background noise. If not given, uses the
            (fixed) value provided to the constructor, or is ignored if a 
            (fixed) noise map has been provided.
        boost_map : jax.numpy.array, optional
            2D map (same dimensions as the model image), that contains
            multiplicative factors for the noise variance. By default 1.
        force_recompute : bool, optional
            If True, forces the use of the fixed noise map. By default False.

        Returns
        -------
        jax.numpy.array
            Noise variance per pixel corresponding to the input image model.
        """
        if not force_recompute and self._noise_map is not None:
            # variance fixed by the noise map
            c_d = self._noise_map**2
        elif self._background_rms is not None:
            # variance computed based on model and fixed background RMS and exposure map
            c_d = self.total_variance(model, self._background_rms, self.exposure_map)
        else:
            # variance computed based on model, given background RMS and fixed exposure map
            c_d = self.total_variance(model, background_rms, self.exposure_map)
        return self.global_boost_map * boost_map * c_d
    
    def _reset_cache(self):
        if hasattr(self, '_C_D'):
            delattr(self, '_C_D')


    @staticmethod
    def total_variance(flux, background_rms, exposure_map):
        """Computes the diagonal of the data covariance matrix, i.e. the noise variance.

        Notes:
        - the exposure map must be positive definite. Values that deviate too much from the mean exposure time will be
            given a lower limit to not under-predict the Poisson component of the noise.
        - the flux must be positive semi-definite for the Poisson noise estimate.
            Values < 0 (Possible after mean subtraction) will not have a Poisson component in their noise estimate.

        Parameters
        ----------
        flux : jax.numpy.array
            Pixels containing the flux from which the shot (Poisson) noise can be estimated.
        background_rms : float
            Standard deviation of the background noise.
        exposure_map : float or jax.numpy.array
            Global exposure time or exposure time per pixel.

        Returns
        -------
        jax.numpy.array
            Noise variance (background + flux-dependent shot noise) per pixel.
        """
        sigma2_bkg = background_rms**2
        if exposure_map is not None:
            flux_pos = jnp.maximum(0, flux)
            sigma2_poisson = flux_pos / exposure_map
            sigma2_tot = sigma2_bkg + sigma2_poisson
        else:
            sigma2_tot = sigma2_bkg * jnp.ones_like(flux)
        return sigma2_tot
